import io
import pkgutil

from PIL import Image
from matplotlib.figure import Figure
from matplotlib.image import FigureImage

from .. import __name__ as _pkg_name_


def add_watermark(fig: Figure) -> FigureImage:
    img_data = pkgutil.get_data(_pkg_name_, 'data/tdy-watermark.png')
    img = Image.open(io.BytesIO(img_data))

    return fig.figimage(img, xo=0, yo=0, zorder=10, alpha=0.25)



from pathlib import Path, PurePath
from typing import Optional, Tuple, Union

import numpy as np
import numpy.typing as npt
from matplotlib.figure import Figure

from ..graphs import draw_3d_locs, draw_occurrences, draw_histogram


def reconstruct_results(fig1: Figure, fig2: Figure, fig3: Figure,
                        locs3d_data_or_file: Union[Path, npt.NDArray[float]],
                        max_lateral_err: Optional[float] = None,
                        min_view_count: Optional[int] = None
                        ) -> Tuple[Figure, Figure, Figure]:
    """Generates 3 final figures with results from stored 3D localisations.

    Args:
        fig3 (Figure): A figure for occurrences.
        fig2 (Figure): A figure for histogram.
        fig1 (Figure): A figure for 3D view.
        locs3d_data_or_file (Path or npt.NDArray[float]):
            A CSV file path or a numpy array with 3D localisations.
        max_lateral_err (Optional[float]):
            Max. lateral error to show (in microns).
        min_view_count(Optional[int]):
            Shown only points with given min. number of views.

    Returns:
        Returns a tuple with 3 figures: occurrences, histogram and 3D view.
    """

    if isinstance(locs3d_data_or_file, np.ndarray):
        locs_3d = locs3d_data_or_file
    elif isinstance(locs3d_data_or_file, PurePath):
        locs_3d = np.genfromtxt(locs3d_data_or_file, delimiter=',', dtype=float)
    else:
        raise TypeError('Unsupported argument type with 3D localisations')

    xyz = locs_3d[:, 0:3]  # X, Y, Z
    lateral_err = locs_3d[:, 3]  # Fitting error in X and Y (in microns)
    axial_err = locs_3d[:, 4]  # Fitting error in Z (in microns)
    view_count = locs_3d[:, 5]  # Number of views used to fit the localisation
    photons = locs_3d[:, 6]  # Number of photons in fit
    # frames = locs_3d[:, 7]  # Frames

    keep = np.logical_and(
        (lateral_err < max_lateral_err) if max_lateral_err is not None else True,
        (view_count > min_view_count) if min_view_count is not None else True)

    draw_occurrences(fig1, lateral_err[keep], axial_err[keep], photons[keep])
    draw_histogram(fig2, photons[keep], axial_err[keep])
    draw_3d_locs(fig3, xyz[keep])

    return fig1, fig2, fig3



import tkinter as tk
from idlelib.tooltip import Hovertip
from tkinter import ttk
from typing import Callable, Optional

import numpy as np

from .app_model import AppModel
from .cfg_dialog import CfgDialog


# pylint: disable=too-many-instance-attributes
class CorrectCfgDialog(CfgDialog):

    def __init__(self, parent, model: AppModel, title: Optional[str] = None,
                 process_cb: Optional[Callable[[bool], None]] = None):
        if title is None:
            title = 'Correction Settings'

        self._ui_frames_min: Optional[ttk.Entry] = None
        self._var_frames_min = tk.StringVar(
            value=str(model.cfg.fit_params_aberration.frame_min))
        self._ui_frames_min_tip: Optional[Hovertip] = None
        self._ui_frames_max: Optional[ttk.Entry] = None
        self._var_frames_max = tk.StringVar(
            value=str(model.cfg.fit_params_aberration.frame_max))
        self._ui_frames_max_tip: Optional[Hovertip] = None

        self._ui_disparity_max: Optional[ttk.Entry] = None
        self._var_disparity_max = tk.StringVar(
            value=str(model.cfg.fit_params_aberration.disparity_max))
        self._ui_disparity_max_tip: Optional[Hovertip] = None

        self._ui_disparity_step: Optional[ttk.Entry] = None
        self._var_disparity_step = tk.StringVar(
            value=str(model.cfg.fit_params_aberration.disparity_step))
        self._ui_disparity_step_tip: Optional[Hovertip] = None

        self._ui_dist_search: Optional[ttk.Entry] = None
        self._var_dist_search = tk.StringVar(
            value=str(model.cfg.fit_params_aberration.dist_search))
        self._ui_dist_search_tip: Optional[Hovertip] = None

        self._ui_angle_tolerance: Optional[ttk.Entry] = None
        self._var_angle_tolerance = tk.StringVar(
            value=str(model.cfg.fit_params_aberration.angle_tolerance))
        self._ui_angle_tolerance_tip: Optional[Hovertip] = None

        self._ui_threshold: Optional[ttk.Entry] = None
        self._var_threshold = tk.StringVar(
            value=str(model.cfg.fit_params_aberration.threshold))
        self._ui_threshold_tip: Optional[Hovertip] = None

        self._ui_min_views: Optional[ttk.Entry] = None
        self._var_min_views = tk.StringVar(
            value=str(model.cfg.fit_params_aberration.min_views))
        self._ui_min_views_tip: Optional[Hovertip] = None

        self._ui_axial_window: Optional[ttk.Entry] = None
        self._var_axial_window = tk.StringVar(
            value=str(model.cfg.aberration_params.axial_window))
        self._ui_axial_window_tip: Optional[Hovertip] = None

        self._ui_photon_threshold: Optional[ttk.Entry] = None
        self._var_photon_threshold = tk.StringVar(
            value=str(model.cfg.aberration_params.photon_threshold))
        self._ui_photon_threshold_tip: Optional[Hovertip] = None

        super().__init__(parent, model, title, process_cb=process_cb)

    # noinspection PyProtectedMember, PyBroadException
    # pylint: disable=protected-access
    def body(self, master) -> tk.BaseWidget:
        ui_tab = ttk.Frame(master)
        row = 0

        try:
            nrs = self.model.lfl.locs_2d[:, 0]
            limits = (f'\n\nFor current data'
                      f' MIN={int(np.min(nrs))}, MAX={int(np.max(nrs))}')
        except BaseException:
            limits = ''
        #
        ui_frames_lbl = ttk.Label(ui_tab, text='Frame numbers:', anchor=tk.E)
        Hovertip(ui_frames_lbl,
                 text=(self.model.cfg._fit_params_aberration_doc
                       + '\nA range of frame numbers included in correction.'
                       + limits))
        self._ui_frames_min = ttk.Entry(
            ui_tab, textvariable=self._var_frames_min)
        self._ui_frames_min_tip = Hovertip(
            self._ui_frames_min,
            text=(self.model.cfg._fit_params_aberration_doc + '\n'
                  + self.model.cfg._fit_params_aberration_frame_min_doc
                  + limits))
        self.ui_widgets.append(self._ui_frames_min)
        self._ui_frames_max = ttk.Entry(
            ui_tab, textvariable=self._var_frames_max)
        self._ui_frames_max_tip = Hovertip(
            self._ui_frames_max,
            text=(self.model.cfg._fit_params_aberration_doc + '\n'
                  + self.model.cfg._fit_params_aberration_frame_max_doc
                  + limits))
        self.ui_widgets.append(self._ui_frames_max)
        #
        ui_frames_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_frames_min.grid(column=1, row=row, sticky=tk.EW)
        self._ui_frames_max.grid(column=2, row=row, sticky=tk.EW)
        row += 1

        ui_disparity_max_lbl = ttk.Label(ui_tab, text='Max. disparity:', anchor=tk.E)
        Hovertip(ui_disparity_max_lbl,
                 text=(self.model.cfg._fit_params_aberration_doc + '\n'
                       + self.model.cfg._fit_params_aberration_disparity_max_doc))
        self._ui_disparity_max = ttk.Entry(
            ui_tab, textvariable=self._var_disparity_max)
        self._ui_disparity_max_tip = Hovertip(
            self._ui_disparity_max,
            text=(self.model.cfg._fit_params_aberration_doc + '\n'
                  + self.model.cfg._fit_params_aberration_disparity_max_doc))
        self.ui_widgets.append(self._ui_disparity_max)
        ui_disparity_max_unit = ttk.Label(ui_tab, text='\u00B5m')  # microns
        #
        ui_disparity_max_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_disparity_max.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_disparity_max_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_disparity_step_lbl = ttk.Label(ui_tab, text='Disparity step:', anchor=tk.E)
        Hovertip(ui_disparity_step_lbl,
                 text=(self.model.cfg._fit_params_aberration_doc + '\n'
                       + self.model.cfg._fit_params_aberration_disparity_step_doc))
        self._ui_disparity_step = ttk.Entry(
            ui_tab, textvariable=self._var_disparity_step)
        self._ui_disparity_step_tip = Hovertip(
            self._ui_disparity_step,
            text=(self.model.cfg._fit_params_aberration_doc + '\n'
                  + self.model.cfg._fit_params_aberration_disparity_step_doc))
        self.ui_widgets.append(self._ui_disparity_step)
        ui_disparity_step_unit = ttk.Label(ui_tab, text='\u00B5m')  # microns
        #
        ui_disparity_step_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_disparity_step.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_disparity_step_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_dist_search_lbl = ttk.Label(ui_tab, text='Search distance:', anchor=tk.E)
        Hovertip(ui_dist_search_lbl,
                 text=(self.model.cfg._fit_params_aberration_doc + '\n'
                       + self.model.cfg._fit_params_aberration_dist_search_doc))
        self._ui_dist_search = ttk.Entry(
            ui_tab, textvariable=self._var_dist_search)
        self._ui_dist_search_tip = Hovertip(
            self._ui_dist_search,
            text=(self.model.cfg._fit_params_aberration_doc + '\n'
                  + self.model.cfg._fit_params_aberration_dist_search_doc))
        self.ui_widgets.append(self._ui_dist_search)
        ui_dist_search_unit = ttk.Label(ui_tab, text='\u00B5m')  # microns
        #
        ui_dist_search_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_dist_search.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_dist_search_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_angle_tolerance_lbl = ttk.Label(ui_tab, text='Angle tolerance:', anchor=tk.E)
        Hovertip(ui_angle_tolerance_lbl,
                 text=(self.model.cfg._fit_params_aberration_doc + '\n'
                       + self.model.cfg._fit_params_aberration_angle_tolerance_doc))
        self._ui_angle_tolerance = ttk.Entry(
            ui_tab, textvariable=self._var_angle_tolerance)
        self._ui_angle_tolerance_tip = Hovertip(
            self._ui_angle_tolerance,
            text=(self.model.cfg._fit_params_aberration_doc + '\n'
                  + self.model.cfg._fit_params_aberration_angle_tolerance_doc))
        self.ui_widgets.append(self._ui_angle_tolerance)
        ui_angle_tolerance_unit = ttk.Label(ui_tab, text='deg')
        #
        ui_angle_tolerance_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_angle_tolerance.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_angle_tolerance_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_threshold_lbl = ttk.Label(ui_tab, text='Threshold:', anchor=tk.E)
        Hovertip(ui_threshold_lbl,
                 text=(self.model.cfg._fit_params_aberration_doc + '\n'
                       + self.model.cfg._fit_params_aberration_threshold_doc))
        self._ui_threshold = ttk.Entry(
            ui_tab, textvariable=self._var_threshold)
        self._ui_threshold_tip = Hovertip(
            self._ui_threshold,
            text=(self.model.cfg._fit_params_aberration_doc + '\n'
                  + self.model.cfg._fit_params_aberration_threshold_doc))
        self.ui_widgets.append(self._ui_threshold)
        ui_threshold_unit = ttk.Label(ui_tab, text='\u00B5m')  # microns
        #
        ui_threshold_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_threshold.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_threshold_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_min_views_lbl = ttk.Label(ui_tab, text='Min. views:', anchor=tk.E)
        Hovertip(ui_min_views_lbl,
                 text=(self.model.cfg._fit_params_aberration_doc + '\n'
                       + self.model.cfg._fit_params_aberration_min_views_doc))
        self._ui_min_views = ttk.Entry(
            ui_tab, textvariable=self._var_min_views)
        self._ui_min_views_tip = Hovertip(
            self._ui_min_views,
            text=(self.model.cfg._fit_params_aberration_doc + '\n'
                  + self.model.cfg._fit_params_aberration_min_views_doc))
        self.ui_widgets.append(self._ui_min_views)
        #
        ui_min_views_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_min_views.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        ui_axial_window_lbl = ttk.Label(ui_tab, text='Axial window:', anchor=tk.E)
        Hovertip(ui_axial_window_lbl,
                 text=(self.model.cfg._aberration_params_doc + '\n'
                       + self.model.cfg._aberration_params_axial_window_doc))
        self._ui_axial_window = ttk.Entry(
            ui_tab, textvariable=self._var_axial_window)
        self._ui_axial_window_tip = Hovertip(
            self._ui_axial_window,
            text=(self.model.cfg._aberration_params_doc + '\n'
                  + self.model.cfg._aberration_params_axial_window_doc))
        self.ui_widgets.append(self._ui_axial_window)
        ui_axial_window_unit = ttk.Label(ui_tab, text='\u00B5m')  # microns
        #
        ui_axial_window_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_axial_window.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        ui_axial_window_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_photon_threshold_lbl = ttk.Label(ui_tab, text='Photon threshold:', anchor=tk.E)
        Hovertip(ui_photon_threshold_lbl,
                 text=(self.model.cfg._aberration_params_doc + '\n'
                       + self.model.cfg._aberration_params_photon_threshold_doc))
        self._ui_photon_threshold = ttk.Entry(
            ui_tab, textvariable=self._var_photon_threshold)
        self._ui_photon_threshold_tip = Hovertip(
            self._ui_photon_threshold,
            text=(self.model.cfg._aberration_params_doc + '\n'
                  + self.model.cfg._aberration_params_photon_threshold_doc))
        self.ui_widgets.append(self._ui_photon_threshold)
        #
        ui_photon_threshold_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_photon_threshold.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        # Final padding and layout
        ui_tab.columnconfigure(index=1, weight=1)
        ui_tab.columnconfigure(index=2, weight=1)
        self.model.grid_slaves_configure(ui_tab, padx=1, pady=1)

        ui_tab.pack(anchor=tk.NW, fill=tk.X, expand=True, padx=5, pady=5)

        return self._ui_frames_max  # Control that gets initial focus

    # pylint: disable=too-many-return-statements
    def validate(self) -> bool:
        if not self.is_int(self._var_frames_min.get()):
            self.initial_focus = self._ui_frames_min
            self._ui_frames_min_tip.showtip()
            return False
        if not self.is_int(self._var_frames_max.get()):
            self.initial_focus = self._ui_frames_max
            self._ui_frames_max_tip.showtip()
            return False
        frames_min = int(self._var_frames_min.get())
        frames_max = int(self._var_frames_max.get())
        if frames_min > frames_max:
            self.initial_focus = self._ui_frames_max
            self._ui_frames_max_tip.showtip()
            return False
        if not self.is_float(self._var_disparity_max.get()):
            self.initial_focus = self._ui_disparity_max
            self._ui_disparity_max_tip.showtip()
            return False
        if not self.is_float(self._var_disparity_step.get()):
            self.initial_focus = self._ui_disparity_step
            self._ui_disparity_step_tip.showtip()
            return False
        if not self.is_float(self._var_dist_search.get()):
            self.initial_focus = self._ui_dist_search
            self._ui_dist_search_tip.showtip()
            return False
        if not self.is_float(self._var_angle_tolerance.get()):
            self.initial_focus = self._ui_angle_tolerance
            self._ui_angle_tolerance_tip.showtip()
            return False
        if not self.is_float(self._var_threshold.get()):
            self.initial_focus = self._ui_threshold
            self._ui_threshold_tip.showtip()
            return False
        if not self.is_int(self._var_min_views.get()):
            self.initial_focus = self._ui_min_views
            self._ui_min_views_tip.showtip()
            return False

        if not self.is_float(self._var_axial_window.get()):
            self.initial_focus = self._ui_axial_window
            self._ui_axial_window_tip.showtip()
            return False
        if not self.is_int(self._var_photon_threshold.get()):
            self.initial_focus = self._ui_photon_threshold
            self._ui_photon_threshold_tip.showtip()
            return False

        return super().validate()

    def process(self) -> None:
        self.model.cfg.fit_params_aberration.frame_min = (
            int(self._var_frames_min.get()))
        self.model.cfg.fit_params_aberration.frame_max = (
            int(self._var_frames_max.get()))
        self.model.cfg.fit_params_aberration.disparity_max = (
            float(self._var_disparity_max.get()))
        self.model.cfg.fit_params_aberration.disparity_step = (
            float(self._var_disparity_step.get()))
        self.model.cfg.fit_params_aberration.dist_search = (
            float(self._var_dist_search.get()))
        self.model.cfg.fit_params_aberration.angle_tolerance = (
            float(self._var_angle_tolerance.get()))
        self.model.cfg.fit_params_aberration.threshold = (
            float(self._var_threshold.get()))
        self.model.cfg.fit_params_aberration.min_views = (
            int(self._var_min_views.get()))

        self.model.cfg.aberration_params.axial_window = (
            float(self._var_axial_window.get()))
        self.model.cfg.aberration_params.photon_threshold = (
            int(self._var_photon_threshold.get()))

        # Used the same value from cfg.fit_params_aberration
        self.model.cfg.aberration_params.min_views = int(self._var_min_views.get())

        super().process()



from typing import Optional

import numpy as np
import numpy.typing as npt
from matplotlib import cm
from matplotlib import colors
from matplotlib.axes import Axes
from matplotlib.colorbar import Colorbar
from matplotlib.figure import Figure
from matplotlib.patches import Circle

from ..graphs import add_watermark


def draw_locs(fig: Figure,
              xy: npt.NDArray[float],
              lens_idx: npt.NDArray[float],
              lens_centres: Optional[npt.NDArray[float]] = None,
              mla_centre: Optional[npt.NDArray[float]] = None,
              bfp_radius: Optional[float] = None,
              set_default_size: bool = True
              ) -> Figure:
    fig.clear(True)
    if set_default_size:
        fig.set_size_inches(6, 5)
    fig.set_layout_engine('tight')

    ax: Axes = fig.add_subplot()
    ax.set_xlabel(r'X [$\mu$m]')
    ax.set_ylabel(r'Y [$\mu$m]')

    lens_idx_uni = np.unique(lens_idx).astype(int)

    cbar_lenses = lens_idx_uni
    cbar_labels = [str(i) for i in cbar_lenses]
    cbar_bounds = np.concatenate((cbar_lenses, [cbar_lenses[-1] + 1])) - 0.5
    cbar_ticks = (cbar_bounds[1:] + cbar_bounds[:-1]) / 2
    cbar_norm = colors.BoundaryNorm(boundaries=cbar_bounds, ncolors=256)
    cbar: Colorbar = fig.colorbar(
        cm.ScalarMappable(norm=cbar_norm), ax=ax,
        boundaries=cbar_bounds, values=cbar_lenses)
    cbar.ax.tick_params(which='both', size=0)  # Hide ticks
    cbar.set_ticks(cbar_ticks.tolist(), labels=cbar_labels)
    cbar.set_label('Lens index')

    if bfp_radius is not None and mla_centre is not None:
        bfp = Circle(xy=(float(mla_centre[0]), float(mla_centre[1])),
                     radius=bfp_radius, fill=False,
                     edgecolor='purple', alpha=0.25)
        ax.add_patch(bfp)

    ax.scatter(xy[:, 0], xy[:, 1], s=1, c=lens_idx, marker=',', lw=0)

    if mla_centre is not None:
        ax.scatter(mla_centre[0], mla_centre[1],
                   s=20, c='red', marker='x')

    if lens_centres is not None:
        lens_centres_uni = lens_centres[lens_idx_uni, :]
        ax.scatter(lens_centres_uni[:, 0], lens_centres_uni[:, 1],
                   s=3, c='tomato')
        for i in lens_idx_uni:
            ax.annotate(str(i),
                        xy=(lens_centres[i, 0], lens_centres[i, 1]),
                        xycoords='data',
                        xytext=(5.0, 2.5),
                        alpha=0.75,
                        textcoords='offset points',
                        size=10)

    ax.set_aspect('equal')
    add_watermark(fig)
    return fig



import tkinter as tk
from idlelib.tooltip import Hovertip
from tkinter import ttk
from typing import Callable, Optional

import numpy as np

import smlfm
from .app_model import AppModel
from .cfg_dialog import CfgDialog
from .consts import READONLY


class FilterCfgDialog(CfgDialog):

    def __init__(self, parent, model: AppModel, title: Optional[str] = None,
                 process_cb: Optional[Callable[[bool], None]] = None):
        if title is None:
            title = 'Model & Filtering Settings'

        self._ui_filter_lenses: Optional[ttk.Checkbutton] = None
        self._var_filter_lenses = tk.BooleanVar(value=model.cfg.filter_lenses)

        self._ui_filter_rhos: Optional[ttk.Checkbutton] = None
        self._var_filter_rhos = tk.BooleanVar(
            value=model.cfg.filter_rhos is not None)
        self._ui_filter_rhos_min: Optional[ttk.Entry] = None
        self._var_filter_rhos_min = tk.StringVar()
        self._ui_filter_rhos_min_tip: Optional[Hovertip] = None
        self._ui_filter_rhos_max: Optional[ttk.Entry] = None
        self._var_filter_rhos_max = tk.StringVar()
        self._ui_filter_rhos_max_tip: Optional[Hovertip] = None
        if model.cfg.filter_rhos is not None:
            self._var_filter_rhos_min.set(str(model.cfg.filter_rhos[0]))
            self._var_filter_rhos_max.set(str(model.cfg.filter_rhos[1]))

        self._ui_filter_spot_sizes: Optional[ttk.Checkbutton] = None
        self._var_filter_spot_sizes = tk.BooleanVar(
            value=model.cfg.filter_spot_sizes is not None)
        self._ui_filter_spot_sizes_min: Optional[ttk.Entry] = None
        self._var_filter_spot_sizes_min = tk.StringVar()
        self._ui_filter_spot_sizes_min_tip: Optional[Hovertip] = None
        self._ui_filter_spot_sizes_max: Optional[ttk.Entry] = None
        self._var_filter_spot_sizes_max = tk.StringVar()
        self._ui_filter_spot_sizes_max_tip: Optional[Hovertip] = None
        if model.cfg.filter_spot_sizes is not None:
            self._var_filter_spot_sizes_min.set(str(model.cfg.filter_spot_sizes[0]))
            self._var_filter_spot_sizes_max.set(str(model.cfg.filter_spot_sizes[1]))

        self._ui_filter_photons: Optional[ttk.Checkbutton] = None
        self._var_filter_photons = tk.BooleanVar(
            value=model.cfg.filter_photons is not None)
        self._ui_filter_photons_min: Optional[ttk.Entry] = None
        self._var_filter_photons_min = tk.StringVar()
        self._ui_filter_photons_min_tip: Optional[Hovertip] = None
        self._ui_filter_photons_max: Optional[ttk.Entry] = None
        self._var_filter_photons_max = tk.StringVar()
        self._ui_filter_photons_max_tip: Optional[Hovertip] = None
        if model.cfg.filter_photons is not None:
            self._var_filter_photons_min.set(str(model.cfg.filter_photons[0]))
            self._var_filter_photons_max.set(str(model.cfg.filter_photons[1]))

        self._ui_alpha_model: Optional[ttk.Combobox] = None
        self._var_alpha_model = tk.StringVar(value=model.cfg.alpha_model.name)

        super().__init__(parent, model, title, process_cb=process_cb)

    # noinspection PyProtectedMember, PyBroadException
    # pylint: disable=protected-access
    def body(self, master) -> tk.BaseWidget:
        ui_tab = ttk.Frame(master)
        row = 0

        self._ui_filter_lenses = ttk.Checkbutton(
            ui_tab, onvalue=True, offvalue=False,
            text='Filter lenses outside back focal plane',
            variable=self._var_filter_lenses)
        Hovertip(self._ui_filter_lenses, text=self.model.cfg._filter_lenses_doc)
        self.ui_widgets.append(self._ui_filter_lenses)
        #
        self._ui_filter_lenses.grid(column=0, row=row, sticky=tk.EW, columnspan=3)
        row += 1

        try:
            uv = self.model.lfl.locs_2d[:, 1:3]
            rho = np.sqrt(np.sum(uv ** 2, axis=1))
            limits = f'\n\nFor current data MIN={np.min(rho)}, MAX={np.max(rho)}'
        except BaseException:
            limits = ''
        #
        ui_filter_rhos = ttk.Checkbutton(
            ui_tab, onvalue=True, offvalue=False, text='Filter rhos:',
            variable=self._var_filter_rhos,
            command=lambda: self.enable(True))
        Hovertip(ui_filter_rhos,
                 text=self.model.cfg._filter_rhos_doc + limits)
        self.ui_widgets.append(ui_filter_rhos)
        self._ui_filter_rhos_min = ttk.Entry(
            ui_tab, textvariable=self._var_filter_rhos_min)
        self._ui_filter_rhos_min_tip = Hovertip(
            self._ui_filter_rhos_min,
            text=self.model.cfg._filter_rhos_doc + limits)
        self.ui_widgets.append(self._ui_filter_rhos_min)
        self._ui_filter_rhos_max = ttk.Entry(
            ui_tab, textvariable=self._var_filter_rhos_max)
        self._ui_filter_rhos_max_tip = Hovertip(
            self._ui_filter_rhos_max,
            text=self.model.cfg._filter_rhos_doc + limits)
        self.ui_widgets.append(self._ui_filter_rhos_max)
        ui_filter_rhos_unit = ttk.Label(ui_tab, text='[npc]')
        Hovertip(ui_filter_rhos_unit, text='normalised pupil coordinates')
        #
        ui_filter_rhos.grid(column=0, row=row, sticky=tk.EW)
        self._ui_filter_rhos_min.grid(column=1, row=row, sticky=tk.EW)
        self._ui_filter_rhos_max.grid(column=2, row=row, sticky=tk.EW)
        ui_filter_rhos_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        try:
            spot_sizes = self.model.lfl.locs_2d[:, 5]
            limits = (f'\n\nFor current data'
                      f' MIN={np.min(spot_sizes)}, MAX={np.max(spot_sizes)}')
        except BaseException:
            limits = ''
        #
        ui_filter_spot_sizes = ttk.Checkbutton(
            ui_tab, onvalue=True, offvalue=False, text='Filter spot sizes:',
            variable=self._var_filter_spot_sizes,
            command=lambda: self.enable(True))
        Hovertip(ui_filter_spot_sizes,
                 text=self.model.cfg._filter_spot_sizes_doc + limits)
        self.ui_widgets.append(ui_filter_spot_sizes)
        self._ui_filter_spot_sizes_min = ttk.Entry(
            ui_tab, textvariable=self._var_filter_spot_sizes_min)
        self._ui_filter_spot_sizes_min_tip = Hovertip(
            self._ui_filter_spot_sizes_min,
            text=self.model.cfg._filter_spot_sizes_doc + limits)
        self.ui_widgets.append(self._ui_filter_spot_sizes_min)
        self._ui_filter_spot_sizes_max = ttk.Entry(
            ui_tab, textvariable=self._var_filter_spot_sizes_max)
        self._ui_filter_spot_sizes_max_tip = Hovertip(
            self._ui_filter_spot_sizes_max,
            text=self.model.cfg._filter_spot_sizes_doc + limits)
        self.ui_widgets.append(self._ui_filter_spot_sizes_max)
        ui_filter_spot_sizes_unit = ttk.Label(ui_tab, text='\u00B5m')  # microns
        #
        ui_filter_spot_sizes.grid(column=0, row=row, sticky=tk.EW)
        self._ui_filter_spot_sizes_min.grid(column=1, row=row, sticky=tk.EW)
        self._ui_filter_spot_sizes_max.grid(column=2, row=row, sticky=tk.EW)
        ui_filter_spot_sizes_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        try:
            photons = self.model.lfl.locs_2d[:, 7]
            limits = (f'\n\nFor current data'
                      f' MIN={np.min(photons)}, MAX={np.max(photons)}')
        except BaseException:
            limits = ''
        #
        ui_filter_photons = ttk.Checkbutton(
            ui_tab, onvalue=True, offvalue=False, text='Filter photons:',
            variable=self._var_filter_photons,
            command=lambda: self.enable(True))
        Hovertip(ui_filter_photons,
                 text=self.model.cfg._filter_photons_doc + limits)
        self.ui_widgets.append(ui_filter_photons)
        self._ui_filter_photons_min = ttk.Entry(
            ui_tab, textvariable=self._var_filter_photons_min)
        self._ui_filter_photons_min_tip = Hovertip(
            self._ui_filter_photons_min,
            text=self.model.cfg._filter_photons_doc + limits)
        self.ui_widgets.append(self._ui_filter_photons_min)
        self._ui_filter_photons_max = ttk.Entry(
            ui_tab, textvariable=self._var_filter_photons_max)
        self._ui_filter_photons_max_tip = Hovertip(
            self._ui_filter_photons_max,
            text=self.model.cfg._filter_photons_doc + limits)
        self.ui_widgets.append(self._ui_filter_photons_max)
        ui_filter_photons_unit = ttk.Label(ui_tab, text='photons')
        #
        ui_filter_photons.grid(column=0, row=row, sticky=tk.EW)
        self._ui_filter_photons_min.grid(column=1, row=row, sticky=tk.EW)
        self._ui_filter_photons_max.grid(column=2, row=row, sticky=tk.EW)
        ui_filter_photons_unit.grid(column=3, row=row, sticky=tk.W)
        row += 1

        ui_alpha_model_lbl = ttk.Label(ui_tab, text='Alpha model:')
        Hovertip(ui_alpha_model_lbl, text=self.model.cfg._alpha_model_doc)
        self._ui_alpha_model = ttk.Combobox(
            ui_tab, state=READONLY,
            values=[t.name for t in smlfm.Localisations.AlphaModel],
            textvariable=self._var_alpha_model)
        Hovertip(self._ui_alpha_model, text=self.model.cfg._alpha_model_doc)
        self.ui_widgets.append(self._ui_alpha_model)
        #
        ui_alpha_model_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_alpha_model.grid(column=1, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        # Final padding and layout
        ui_tab.columnconfigure(index=1, weight=1)
        ui_tab.columnconfigure(index=2, weight=1)
        self.model.grid_slaves_configure(ui_tab, padx=1, pady=1)

        ui_tab.pack(anchor=tk.NW, fill=tk.X, expand=True, padx=5, pady=5)

        return self._ui_filter_lenses  # Control that gets initial focus

    # pylint: disable=too-many-return-statements
    # pylint: disable=protected-access
    def validate(self) -> bool:
        if self._var_filter_rhos.get():
            if not self.is_float(self._var_filter_rhos_min.get()):
                self.initial_focus = self._ui_filter_rhos_min
                self._ui_filter_rhos_min_tip.showtip()
                return False
            if not self.is_float(self._var_filter_rhos_max.get()):
                self.initial_focus = self._ui_filter_rhos_max
                self._ui_filter_rhos_max_tip.showtip()
                return False
            rhos_min = float(self._var_filter_rhos_min.get())
            rhos_max = float(self._var_filter_rhos_max.get())
            if rhos_min > rhos_max:
                self.initial_focus = self._ui_filter_rhos_max
                self._ui_filter_rhos_max_tip.showtip()
                return False
        if self._var_filter_spot_sizes.get():
            if not self.is_float(self._var_filter_spot_sizes_min.get()):
                self.initial_focus = self._ui_filter_spot_sizes_min
                self._ui_filter_spot_sizes_min_tip.showtip()
                return False
            if not self.is_float(self._var_filter_spot_sizes_max.get()):
                self.initial_focus = self._ui_filter_spot_sizes_max
                self._ui_filter_spot_sizes_max_tip.showtip()
                return False
            spot_sizes_min = float(self._var_filter_spot_sizes_min.get())
            spot_sizes_max = float(self._var_filter_spot_sizes_max.get())
            if spot_sizes_min > spot_sizes_max:
                self.initial_focus = self._ui_filter_spot_sizes_max
                self._ui_filter_spot_sizes_max_tip.showtip()
                return False
        if self._var_filter_photons.get():
            if not self.is_float(self._var_filter_photons_min.get()):
                self.initial_focus = self._ui_filter_photons_min
                self._ui_filter_photons_min_tip.showtip()
                return False
            if not self.is_float(self._var_filter_photons_max.get()):
                self.initial_focus = self._ui_filter_photons_max
                self._ui_filter_photons_max_tip.showtip()
                return False
            photons_min = float(self._var_filter_photons_min.get())
            photons_max = float(self._var_filter_photons_max.get())
            if photons_min > photons_max:
                self.initial_focus = self._ui_filter_photons_max
                self._ui_filter_photons_max_tip.showtip()
                return False

        return super().validate()

    def process(self) -> None:
        self.model.cfg.filter_lenses = self._var_filter_lenses.get()

        if self._var_filter_rhos.get():
            self.model.cfg.filter_rhos = [
                float(self._var_filter_rhos_min.get()),
                float(self._var_filter_rhos_max.get())]
        else:
            self.model.cfg.filter_rhos = None

        if self._var_filter_spot_sizes.get():
            self.model.cfg.filter_spot_sizes = [
                float(self._var_filter_spot_sizes_min.get()),
                float(self._var_filter_spot_sizes_max.get())]
        else:
            self.model.cfg.filter_spot_sizes = None

        if self._var_filter_photons.get():
            self.model.cfg.filter_photons = [
                float(self._var_filter_photons_min.get()),
                float(self._var_filter_photons_max.get())]
        else:
            self.model.cfg.filter_photons = None

        self.model.cfg.alpha_model = smlfm.Localisations.AlphaModel[
            self._var_alpha_model.get()]

        super().process()

    def enable(self, active: bool, change_cursor: bool = True) -> None:
        super().enable(active, change_cursor)

        if active:
            state = tk.NORMAL if self._var_filter_rhos.get() else tk.DISABLED
            self._ui_filter_rhos_min.configure(state=state)
            self._ui_filter_rhos_max.configure(state=state)

            state = tk.NORMAL if self._var_filter_spot_sizes.get() else tk.DISABLED
            self._ui_filter_spot_sizes_min.configure(state=state)
            self._ui_filter_spot_sizes_max.configure(state=state)

            state = tk.NORMAL if self._var_filter_photons.get() else tk.DISABLED
            self._ui_filter_photons_min.configure(state=state)
            self._ui_filter_photons_max.configure(state=state)



import dataclasses
import multiprocessing as mp
from dataclasses import dataclass
from typing import Callable, List, Optional, Tuple

import numpy as np
import numpy.typing as npt


class Fitting:
    """TODO: Add documentation."""

    @dataclass
    class FitParams:
        # A min. frame number.
        frame_min: int
        # A max. frame number.
        frame_max: int
        # A limit of possible disparity between adjacent views,
        # from negative to positive value (in microns).
        disparity_max: float
        # A step used for disparity range (in microns).
        disparity_step: float
        # A window to accept points as belonging to a given Z (in microns).
        dist_search: float
        # A tolerance for selecting with angles around relative UV (in degrees).
        angle_tolerance: float
        # A max. distance of the candidate (in microns).
        threshold: float
        # A min. number of views the candidate is seen to count it.
        min_views: int
        # A calibration factor between optical and physical Z.
        z_calib: Optional[float] = None

    @dataclass
    class AberrationParams:
        # A max. standard error in Z to apply correction (in microns).
        axial_window: float
        # A min. number of photons to apply correction.
        photon_threshold: int
        # A min. number of views the candidate is seen to count it.
        min_views: int

    @dataclass
    class FitData:
        # A frame number.
        frame: int
        # Basically X, Y and uncalibrated Z coordinates in 3D space (3 values).
        model: npt.NDArray[float]
        # A points included in the fit in format from Localisations class
        # (13 columns).
        points: npt.NDArray[float]
        # Total number of photons in the fit.
        photon_count: int
        # An estimated standard error of model values (3 values).
        std_err: npt.NDArray[float]

    @staticmethod
    def light_field_fit(locs_2d: npt.NDArray[float],
                        rho_scaling: float,
                        fit_params: FitParams,
                        abort_event: Optional[mp.Event] = None,
                        progress_func: Optional[Callable[[int, int, int], None]] = None,
                        progress_step: int = 1000,
                        worker_count: int = 0
                        ) -> Tuple[npt.NDArray[float], List[FitData]]:
        """TODO: Add documentation.

        Args:
            locs_2d (npt.NDArray[float]):
                A localisation data in format from Localisations class,
                i.e. 13 columns.
            rho_scaling (float):
                A scaling factor to convert microns in image plane to rho.
            fit_params (FitParams):
                A configuration for fit detection.
            abort_event (Optional[mp.Event]):
                An event to be periodically checked (without blocking).
                If set, the processing is aborted and function returns.
            progress_func (Optional[Callable[[int, int, int], None]]):
                A callback function for reporting progress. The function gets
                three arguments. First is a frame number of currently processed
                frame, the second and third are min. and max. frame numbers
                received via `fit_params`.
            progress_step (int):
                Defines how often is the `progress_func` function called, if set.
                The frames are processed in groups of 100 frames which is also
                the lowest value.
            worker_count (int):
                It is the number of worker processes to run in parallel.
                If it is 0 or less, the number returned by `os.cpu_count()` is used.

        Returns:
            Returns a tuple.

            - fitted_points (npt.NDArray[float]):
              Holds the fitted data in 2D array with following 8 columns:

              - [0] model data X (float, in microns)
              - [1] model data Y (float, in microns)
              - [2] model data Z (float, in microns)
              - [3] lateral error, a mean of standard errors on X,Y
                    (float, in microns)
              - [4] axial error, a standard error on Z (float, in microns)
              - [5] number of views in fit (int, but stored as float)
              - [6] number of photons in fit (float)
              - [7] frame number (int, but stored as float)

            - total_fit (List[FitData]):
              Holds the list of FitData objects so that individually chosen
              points can be inspected.
        """

        min_frame = fit_params.frame_min
        max_frame = fit_params.frame_max
        frame_count = max_frame - min_frame + 1

        if min_frame > max_frame:
            raise ValueError(
                f'min. frame nr. ({min_frame}) > max. frame nr. ({max_frame})')

        frames_per_task = 100
        task_count = int((frame_count + frames_per_task - 1) / frames_per_task)

        if worker_count == 1 or task_count == 1:
            return Fitting._light_field_fit_task(
                locs_2d, rho_scaling, fit_params, abort_event)

        fitted_points = np.empty((0, 8))
        total_fit = []  # List of FitData

        frame = min_frame
        progress_next_frame = min_frame + progress_step

        processes = worker_count if worker_count > 0 else None
        with mp.Pool(processes=processes) as pool:
            procs = [Optional[mp.Process]] * task_count

            for idx in range(task_count):
                fit_params_n = dataclasses.replace(
                    fit_params,
                    frame_min=min_frame + frames_per_task * idx,
                    frame_max=min(min_frame + frames_per_task * (idx + 1), max_frame),
                )
                # Cannot pass the `abort_event` to processes, otherwise following
                # runtime error is raised: "Conditional objects should only be
                # shared between processes through inheritance"
                args = (locs_2d, rho_scaling, fit_params_n, None)
                procs[idx] = pool.apply_async(Fitting._light_field_fit_task, args)

            for idx in range(task_count):
                fitted_points_n, total_fit_n = procs[idx].get()

                if abort_event is not None:
                    if abort_event.is_set():
                        return np.empty((0, 8)), []

                fitted_points = np.row_stack((fitted_points, fitted_points_n))
                total_fit += total_fit_n

                if progress_func is not None:
                    frame = min(frame + frames_per_task, max_frame)
                    if frame >= progress_next_frame or frame == max_frame:
                        progress_func(frame, min_frame, max_frame)
                        progress_next_frame += progress_step

        return fitted_points, total_fit

    @staticmethod
    def _light_field_fit_task(locs_2d: npt.NDArray[float],
                              rho_scaling: float,
                              fit_params: FitParams,
                              abort_event: Optional[mp.Event] = None,
                              progress_func: Optional[Callable[[int, int, int], None]] = None,
                              progress_step: int = 1000
                              ) -> Tuple[npt.NDArray[float], List[FitData]]:
        """A task executed multiple times in separate process to speed up."""

        min_frame = fit_params.frame_min
        max_frame = fit_params.frame_max
        min_views = fit_params.min_views
        fit_threshold = fit_params.threshold
        z_calib = fit_params.z_calib

        fitted_points = np.empty((0, 8))
        total_fit = []  # List of FitData

        progress_next_frame = min_frame + progress_step

        for frame in range(min_frame, max_frame + 1):

            if abort_event is not None:
                if abort_event.is_set():
                    return np.empty((0, 8)), []

            if progress_func is not None:
                if frame >= progress_next_frame or frame == max_frame:
                    progress_func(frame, min_frame, max_frame)
                    progress_next_frame += progress_step

            # Sort potential seeds by central view first
            candidates = locs_2d[locs_2d[:, 0] == frame, :].copy()
            u = candidates[:, 1]  # Read-only, no copy needed
            v = candidates[:, 2]  # Read-only, no copy needed
            loc_cen_sel = np.logical_and(np.abs(u) < 0.1, np.abs(v) < 0.1)
            loc_cen = candidates[loc_cen_sel, :].copy()
            loc_out_sel = ~np.logical_and(u == 0.0, v == 0.0)
            loc_out = candidates[loc_out_sel, :].copy()
            # Then descending sort by intensity column
            loc_cen = loc_cen[loc_cen[:, 7].argsort()[::-1]]
            loc_out = loc_out[loc_out[:, 7].argsort()[::-1]]
            candidates = np.row_stack((loc_cen, loc_out))

            # Loop over localisations until a few left
            reps = 0
            while candidates.shape[0] > min_views and reps < 100:
                reps += 1
                seed = candidates[0, :]
                loc_fit, view_count, _ = Fitting._group_localisations(
                    seed, candidates, fit_params, rho_scaling)
                # If there are not enough views, don't fit and move on
                if view_count < min_views:
                    candidates = candidates[1:, :]  # Remove first row, the seed
                    continue

                # Fit to points
                model, std_err, _ = Fitting._get_backward_model(loc_fit, rho_scaling)
                # Calculate error in model
                dist = np.sqrt(np.sum(std_err**2))
                # If the final fit distance is too big, don't fit and move on
                if dist > fit_threshold:
                    candidates = candidates[1:, :]  # Remove first row, the seed
                    continue

                # If dist threshold met, remove fitted localisations from candidates
                candidates_sel = np.all(candidates[:, None] == loc_fit, axis=-1).any(-1)
                candidates = candidates[~candidates_sel, :]

                # Add to modelled and fitted points
                if loc_fit.size > 0:
                    photon_count = np.sum(loc_fit[:, 7])
                    new_fit_point = np.array([
                        model[0],  # X
                        model[1],  # Y
                        model[2],  # Z (uncalibrated)
                        np.mean(std_err[0:2]),  # Lateral error
                        std_err[2],  # Axial error
                        view_count + 1,  # Number of views
                        photon_count,  # Number of photons in fit
                        frame  # Frame number
                    ])
                    new_fit_data = Fitting.FitData(
                        frame=frame,
                        model=model,
                        points=loc_fit.copy(),
                        photon_count=photon_count,
                        std_err=std_err,
                    )

                    fitted_points = np.row_stack((fitted_points, new_fit_point))
                    total_fit.append(new_fit_data)

        if z_calib is not None:
            fitted_points[:, 2] *= z_calib

        return fitted_points, total_fit

    @staticmethod
    def calculate_view_error(locs_2d: npt.NDArray[float],
                             rho_scaling: float,
                             fit_data: List[FitData],
                             aberration_params: AberrationParams
                             ) -> npt.NDArray[float]:
        """Calculates average fit error.

        Args:
            locs_2d (npt.NDArray[float]):
                A localisation data in format from Localisations class.
            rho_scaling (float):
                A scaling factor to convert microns in image plane to rho.
            fit_data (List[FitData]):
                A list of FitData objects.
            aberration_params (AberrationParams):
                A configuration for aberration detection.

        Returns:
            Returns average fit error, a sort of average aberration, across all
            fitted points (in microns), with following 5 columns:

            * [0] Original U coordinate (in normalised pupil coordinates)
            * [1] Original V coordinate (in normalised pupil coordinates)
            * [2] Correction of X coordinate (in microns)
            * [3] Correction of Y coordinate (in microns)
            * [4] Number of points used for correction
        """

        min_views = aberration_params.min_views
        photon_threshold = aberration_params.photon_threshold
        axial_window = aberration_params.axial_window
        views = np.unique(locs_2d[:, 1:3], axis=0)  # U,V
        view_count = views.shape[0]
        correction = np.zeros((view_count, 5))
        correction[:, 0:2] = views  # U,V
        for fd in fit_data:
            model = fd.model
            points = fd.points
            photon_count = fd.photon_count
            fit_err = Fitting._get_forward_model_error(model, points, rho_scaling)
            if (points.shape[0] > min_views
                    and np.abs(model[2]) < axial_window
                    and photon_count > photon_threshold):
                for j in range(view_count):
                    dum = fit_err[np.logical_and(
                        points[:, 1] == views[j, 0],
                        points[:, 2] == views[j, 1]), :]
                    if dum.size > 0:
                        correction[j, 2:4] += dum[0]  # Only one match at a time
                        correction[j, 4] += 1

        # Avoid divisions by zero, store zero results instead
        x = correction[:, 2]
        y = correction[:, 3]
        c = correction[:, 4]
        correction[:, 2] = np.divide(x, c, out=np.zeros_like(x), where=c != 0)
        correction[:, 3] = np.divide(y, c, out=np.zeros_like(y), where=c != 0)

        return correction

    @staticmethod
    def _get_backward_model(locs: npt.NDArray[float],
                            rho_scaling: float
                            ) -> Tuple[npt.NDArray[float], npt.NDArray[float], float]:
        """TODO: Add documentation.

        Args:
            locs (npt.NDArray[float]):
                A localisation data in format from Localisations class.
            rho_scaling (float):
                A scaling factor to convert microns in image plane to rho.

        Returns:
            Returns a tuple of 3 values.

            - model (npt.NDArray[float]): An ordinary least-squares (OLS)
              estimates of the regression coefficients (3 values).
              Basically X, Y and uncalibrated Z coordinates in 3D space.
            - std_err (npt.NDArray[float]): An estimated standard error of model
              values (3 values).
            - mse (float): An estimated standard deviation of the
              regression error (a mean squared error).
        """

        # These variables are not modified, no need to copy locs data
        u = locs[:, 1]
        v = locs[:, 2]
        x = locs[:, 3] - u / rho_scaling
        y = locs[:, 4] - v / rho_scaling
        alpha_u = locs[:, 10]
        alpha_v = locs[:, 11]

        zeros = np.zeros_like(alpha_u)
        ones = np.ones_like(alpha_u)

        a = np.row_stack((
            np.column_stack((ones, zeros, alpha_u)),
            np.column_stack((zeros, ones, alpha_v))))
        b = np.concatenate((x, y))

        model, residuals, rank, _s = np.linalg.lstsq(a, b, rcond=None)
        mse = np.mean(residuals) / (a.shape[0] - rank)
        covariant_matrix = mse * np.linalg.inv(np.dot(a.T, a))
        std_err = np.sqrt(np.diag(covariant_matrix))

        return model, std_err, mse

    @staticmethod
    def _get_forward_model_error(model: npt.NDArray[float],
                                 points: npt.NDArray[float],
                                 rho_scaling: float
                                 ) -> npt.NDArray[float]:
        """TODO: Add documentation.

        Args:
            model (npt.NDArray[float]):
                A model returned from `_get_backward_model` function (3 values).
            points (npt.NDArray[float]):
                A candidates returned from `_group_localisations` function.
            rho_scaling (float):
                A scaling factor to convert microns in image plane to rho.

        Returns:
            TODO: Add documentation.
        """

        # These variables are not modified, no need to copy the data
        uv = points[:, 1:3]
        xy = points[:, 3:5]
        alpha_uv = points[:, 10:12]

        xy0 = model[0:2]
        z = model[2]

        dist = xy - (uv / rho_scaling) - xy0 - (z * alpha_uv)
        return dist

    @staticmethod
    def _group_localisations(seed: npt.NDArray[float],
                             in_points: npt.NDArray[float],
                             fit_params: FitParams,
                             rho_scaling: float
                             ) -> Tuple[npt.NDArray[float], int, int]:
        """TODO: Add documentation.

        Args:
            seed (npt.NDArray[float]):
                A single row from Localisation data array (13 columns).
            in_points (npt.NDArray[float]):
                A multiple rows from Localisation data array (13 columns).
            fit_params (FitParams):
                A configuration for grouping.
            rho_scaling (float):
                A scaling factor to convert microns in image plane to rho.

        Returns:
            Returns a tuple of following 3 values.

            - candidates (npt.NDArray[float]): A subset of input points which
              are candidates for being related to a single point given as
              `seed`, (13 columns).
            - view_count(int): A number of views.
            - duplicate_count(int): A number of duplicates.
        """

        angle_tol = np.deg2rad(fit_params.angle_tolerance)
        max_disparity = fit_params.disparity_max
        dis_tol = fit_params.dist_search
        dz = fit_params.disparity_step

        # Seed values are not modified, copy not needed
        su = seed[1]
        sv = seed[2]
        sx = seed[3]
        sy = seed[4]

        points = in_points.copy()

        # First, remove self view points
        seed_sel = np.logical_and(points[:, 1] == su, points[:, 2] == sv)
        points = points[~seed_sel, :]

        # Disparity from seed and relative U and V
        du = points[:, 1] - su
        dv = points[:, 2] - sv
        dx = points[:, 3] - sx
        dy = points[:, 4] - sy

        # Angle selection is relative to the UV index of the seed
        # and dx dy on camera pixels.
        angles_uv = np.arctan2(dv, du)
        angles_xy = np.arctan2(dy, dx)
        angle_sel = np.logical_or(
            np.abs(angles_xy - angles_uv) < angle_tol,
            np.abs(angles_xy - angles_uv) > (np.pi - angle_tol))
        points = points[angle_sel, :]

        # ortho_dist = r * np.sin(np.abs(angles_xy - angles_uv))
        # rect_sel = np.abs(ortho_dist) < dis_tol
        # points = points[~rect_sel, :]

        # Select against relative disparity
        du = points[:, 1] - su
        dv = points[:, 2] - sv
        dx = points[:, 3] - sx
        dy = points[:, 4] - sy
        dxy = np.sqrt(dx**2 + dy**2)
        duv = np.sqrt(du**2 + dv**2)  # Rho
        disparity = (dxy - duv / rho_scaling) / duv

        range_z = np.arange(-max_disparity, max_disparity + dz / 2, dz)
        num_points = np.zeros((range_z.size, 1))
        # The num_points will be in integration of the number of points in
        # each disparity gradient range.
        for z_i in range(range_z.size):
            disparity_sel = np.logical_and(
                disparity > (range_z[z_i] - dis_tol),
                disparity <= (range_z[z_i] + dis_tol))
            num_points[z_i] = np.sum(disparity_sel)
        best_z_i = np.argmax(num_points)
        best_z = range_z[best_z_i]
        # Apply Z-selection
        z_sel = np.logical_or(
            disparity <= (best_z - dis_tol),
            disparity > (best_z + dis_tol))
        points = points[~z_sel, :]

        # Get rid of multiple points in single views.
        # This section looks for multiple localisation within a single view
        # that made the selection - we only need one so pick one.
        # This could be done better, I'm sure!
        # I take the one which has best fit. (not all possible combinations
        # checked though, just as they are found).

        # Initialise set of localisations to fit with the seed being used first
        candidates = seed
        view_count = 0
        duplicate_count = 0
        if points.size > 0:
            uv, uv_i, uv_c = np.unique(points[:, 1:3], axis=0,
                                       return_inverse=True, return_counts=True)
            # Generates indices for localisations belonging to a view
            indices = list([])
            for value in uv:
                indices.append(np.where(np.all(points[:, 1:3] == value, axis=1))[0])
            # Add unique points to candidates
            candidates = np.row_stack(
                (candidates, points[np.isin(uv_i, np.nonzero(uv_c == 1))]))
            # Remove indices with uv_c == 1
            indices = [i for i in indices if len(i) > 1]

            view_count = len(uv_c)
            duplicate_count = len(indices)

            # If there is more than one point in a view,
            # take the one which fits the unique points best
            for i in range(duplicate_count):
                view_ind = indices[i]
                j_count = view_ind.shape[0]
                d = np.zeros((j_count, 1))
                for j in range(j_count):
                    locs = np.row_stack((candidates, points[view_ind[j], :]))
                    std_err = Fitting._get_backward_model(locs, rho_scaling)[1]
                    d[j] = np.sqrt(np.sum(std_err**2))
                d_min_j = np.argmin(d)
                candidates = np.row_stack((candidates, points[view_ind[d_min_j], :]))

            # If there are 2 views that share common U or V, reject the grouping
            var_u = np.var(candidates[:, 1], axis=0, ddof=1)
            var_v = np.var(candidates[:, 2], axis=0, ddof=1)
            if view_count == 1 and (var_u < 0.1 or var_v < 0.1):
                candidates = np.array([])
                view_count = 0

        return candidates, view_count, duplicate_count



import tkinter as tk

from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
from matplotlib.figure import Figure


class FigureWindow(tk.Toplevel):

    def __init__(self, figure: Figure, *args, **kwargs):
        title = kwargs.pop('title', None)
        super().__init__(*args, **kwargs)

        if title is not None:
            self.wm_title(title)

        self.figure = figure
        if self.figure.canvas.manager is not None:
            print('WARNING: Tkinter does not like pyplot figures')

        size = self.figure.bbox.size.astype(int)
        self.geometry(f'{size[0]}x{size[1]}')

        self.canvas = FigureCanvasTkAgg(self.figure, master=self)
        self.canvas.draw()

        # pack_toolbar=False makes it easier to use a layout manager later on
        self.toolbar = NavigationToolbar2Tk(self.canvas, window=self,
                                            pack_toolbar=False)
        self.toolbar.update()

        # Packing order is important. Widgets are processed sequentially and if
        # there is no space left, because the window is too small, they are not
        # displayed. The canvas is rather flexible in its size, so we pack it
        # last which makes sure the UI controls are displayed as long as possible.
        self.toolbar.pack(side=tk.BOTTOM, fill=tk.X)
        self.canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH, expand=True)

        # Just hide the window instead of releasing it.
        # Use `destroy()` to release it from memory.
        self.protocol('WM_DELETE_WINDOW', func=self.withdraw)
        self.bind('<Escape>', func=lambda _evt: self.withdraw())

    def refresh(self):
        self.canvas.draw()



import numpy.typing as npt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ..graphs import add_watermark


def draw_histogram(fig: Figure,
                   photons: npt.NDArray[float],
                   axial_err: npt.NDArray[float],
                   bins: int = 40,
                   set_default_size: bool = True
                   ) -> Figure:
    fig.clear(True)
    if set_default_size:
        fig.set_size_inches(6, 5)
    fig.set_layout_engine('tight')

    ax: Axes = fig.add_subplot()
    ax.set_xlabel('Photons')
    ax.set_ylabel('Axial precision [nm]')
    gr = ax.hist2d(photons, axial_err * 1000, bins=bins)

    cbar = fig.colorbar(gr[3], ax=ax)
    cbar.set_label('Number of localisations')

    add_watermark(fig)
    return fig



#!/usr/bin/env python3

import pkgutil
from dataclasses import fields

import pytest

import smlfm


@pytest.mark.parametrize(
    "indents",
    (
        pytest.param(None),
        pytest.param(1),
        pytest.param(4),
    ),
)
def test_config(indents):
    cfg1_dump = pkgutil.get_data(
        smlfm.__name__, 'data/default-config.json').decode()
    cfg1 = smlfm.Config.from_json(cfg1_dump)

    # Deserialization should result in exactly same data types and values
    cfg2_dump = cfg1.to_json(indent=indents)
    cfg2 = smlfm.Config.from_json(cfg2_dump)
    for field in fields(smlfm.Config):
        cfg1_val = getattr(cfg1, field.name)
        cfg2_val = getattr(cfg2, field.name)
        msg = (f'{field.name} (type {field.type}): '
               f'"{cfg1_val}" (type {type(cfg1_val)}) != '
               f'"{cfg2_val}" (type {type(cfg2_val)})')
        if cfg1_val != cfg2_val:
            raise ValueError(f'Value mismatch: {msg}')
        if type(cfg1_val) is not type(cfg2_val):
            raise TypeError(f'Type mismatch: {msg}')



from enum import Enum, unique
from pathlib import Path
from typing import Optional

import numpy as np
import numpy.typing as npt


class LocalisationFile:
    """A wrapper class around localisation files.

    Attributes:
        csv_file (Path): A file name with path to parse.
        file_format (Format): A format of the given file.
        pixel_size (float):
            A pixel size in sample space (in microns), defaults to 1.0.
        data (npt.NDArray[float]):
            A 2D array filled with the data from given file.
            It contains the localised maxima in rows, where each has 8 columns.

            * [0] frame number (int)
            * [1] X coordinate (in microns)
            * [2] Y coordinate (in microns)
            * [3] sigma X (in microns)
            * [4] sigma Y (in microns)
            * [5] intensity (in photons)
            * [6] background intensity (in photons)
            * [7] precision or uncertainty (in microns)

            The `PEAKFIT` file format has data in columns 1 to 4 in pixels,
            not microns. The `scale_peakfit_data()` function should be called
            with a scale factor that converts the value from pixels to microns.
    """

    @unique
    class Format(Enum):
        PEAKFIT = 1
        THUNDERSTORM = 2
        PICASSO = 3

    def __init__(self,
                 csv_file: Path,
                 file_format: Format):
        """Constructs the localisation file wrapper.

        Args:
            csv_file (Path): A file name with path to parse.
            file_format (Format): A format of the given file.
        """

        self.csv_file = csv_file
        self.file_format = file_format
        self.pixel_size: float = 1.0
        self.data: Optional[npt.NDArray[float]] = None

        self._raw_data: Optional[npt.NDArray[float]] = None

    def read(self) -> None:
        """Parses the CSV file."""

        self.pixel_size = 1.0
        self.data = None
        self._raw_data = None

        if self.file_format == LocalisationFile.Format.PICASSO:
            id_frame = 1       # column index with frame number
            id_x = 2           # column index with X coordinates (in nm)
            id_y = 3           # column index with Y coordinates (in nm)
            id_sigma_x = 4     # column index with sigma X (in nm)
            id_sigma_y = 4     # column index with sigma Y (in nm)
            id_intensity = 5   # column index with intensity (in photons)
            id_background = 6  # column index with background (in photons)
            id_precision = 8   # column index with precision or uncertainty (in nm)
            min_columns = 9
            scale_xy = 1000    # scale coordinates to microns
            scale_pr = 1000    # scale precision to microns
        elif self.file_format == LocalisationFile.Format.THUNDERSTORM:
            id_frame = 0       # column index with frame number
            id_x = 1           # column index with X coordinates (in nm)
            id_y = 2           # column index with Y coordinates (in nm)
            id_sigma_x = 3     # column index with sigma X (in nm)
            id_sigma_y = 3     # column index with sigma Y (in nm)
            id_intensity = 4   # column index with intensity (in photons)
            id_background = 5  # column index with background (in photons)
            id_precision = 6   # column index with precision or uncertainty (in nm)
            min_columns = 7
            scale_xy = 1000    # scale coordinates to microns
            scale_pr = 1000    # scale precision to microns
        elif self.file_format == LocalisationFile.Format.PEAKFIT:
            # Reset scale factor, must be applied again on new data
            self.pixel_size = 1.0
            id_frame = 0       # column index with frame number
            id_x = 9           # column index with X coordinates (in pixels)
            id_y = 10          # column index with Y coordinates (in pixels)
            id_sigma_x = 12    # column index with sigma X (in pixels)
            id_sigma_y = 12    # column index with sigma Y (in pixels)
            id_intensity = 8   # column index with intensity (in photons)
            id_background = 7  # column index with background (in photons)
            id_precision = 13  # column index with precision or uncertainty (in nm)
            min_columns = 14
            scale_xy = 1       # scale coordinates to microns (to be set yet)
            scale_pr = 1000    # scale precision to microns
        else:
            raise ValueError(f'Unsupported localisation file format with value'
                             f' {self.file_format}')

        csv_header_rows = 0
        csv_delimiter = None  # Default delimiter (whitespaces)
        max_line_len = 16384
        with open(self.csv_file, 'r', encoding='utf-8') as f:
            while True:
                line = f.readline(max_line_len)
                if not line:
                    break  # No data in file
                if line[-1] != '\n' and len(line) == max_line_len:
                    raise ValueError(f'Row {csv_header_rows + 1} is too long')
                if line[0].isdigit():
                    # First row with data, test comma delimiter
                    if len(line.split(',')) >= min_columns:
                        csv_delimiter = ','
                    break
                csv_header_rows += 1

        raw_data = np.genfromtxt(self.csv_file, delimiter=csv_delimiter,
                                 dtype=float, skip_header=csv_header_rows)

        if raw_data.shape[0] <= 0:
            raise ValueError('No valid data found')
        if raw_data.shape[1] < min_columns:
            raise ValueError(f'Insufficient number of columns for selected format,'
                             f'required min. {min_columns}, found {raw_data.shape[1]}')

        self._raw_data = np.empty((raw_data.shape[0], 8))
        self._raw_data.fill(np.nan)
        self._raw_data[:, 0] = raw_data[:, id_frame]
        self._raw_data[:, 1] = raw_data[:, id_x] / scale_xy
        self._raw_data[:, 2] = raw_data[:, id_y] / scale_xy
        self._raw_data[:, 3] = raw_data[:, id_sigma_x] / scale_xy
        self._raw_data[:, 4] = raw_data[:, id_sigma_y] / scale_xy
        self._raw_data[:, 5] = raw_data[:, id_intensity]
        self._raw_data[:, 6] = raw_data[:, id_background]
        self._raw_data[:, 7] = raw_data[:, id_precision] / scale_pr

        self.data = self._raw_data.copy()

    def scale_peakfit_data(self,
                           pixel_size: float
                           ) -> None:
        """Scales the raw X,Y coordinates and sigmas to microns.

        The scaling is done for `PEAKFIT` file format only.
        It is a no-op for other formats.

        The scale is applied to original raw data, thus it reverts any
        previously set scale.

        Args:
            pixel_size (float): A pixel size in sample space (in microns).
        """

        if self.file_format == LocalisationFile.Format.PEAKFIT:
            if self.pixel_size != pixel_size:
                self.pixel_size = pixel_size
                # Apply new scale to X,Y coordinates and sigmas
                self.data[:, 1:5] = self._raw_data[:, 1:5] * self.pixel_size



import pkgutil
import tkinter as tk
import traceback as tb
from idlelib.tooltip import Hovertip
from pathlib import Path
from threading import Thread
from tkinter import filedialog, messagebox, ttk
from typing import Optional

import smlfm

from .app_model import AppModel, IStage
from .consts import COMMAND, IMAGE, TEXT, TEXTVARIABLE, VARIABLE, StageType
from .config_cfg_dialog import ConfigCfgDialog


# pylint: disable=too-many-ancestors
class ConfigFrame(ttk.Frame, IStage):

    def __init__(self, master, model: AppModel, *args, **kwargs):
        super().__init__(master, *args, **kwargs)

        self._stage_type = StageType.CONFIG
        self._stage_type_next = StageType.CSV

        self._model = model

        # Controls
        self._ui_frm_lbl = ttk.Label(self)
        self._ui_frm = ttk.LabelFrame(self, labelwidget=self._ui_frm_lbl)

        self._ui_tab = ttk.Frame(self._ui_frm)
        self._ui_file_lbl = ttk.Label(self._ui_tab)
        self._ui_file = ttk.Entry(self._ui_tab)
        self._ui_file_tip = Hovertip(self._ui_file, text='')
        self._ui_file_btn = ttk.Button(self._ui_tab)

        self._ui_buttons = ttk.Frame(self._ui_tab)
        self._ui_open = ttk.Button(self._ui_buttons)
        self._ui_save_as = ttk.Button(self._ui_buttons)

        self._ui_summary = ttk.Frame(self._ui_frm)
        self._ui_settings = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_settings_tip = Hovertip(self._ui_settings, text='')
        self._ui_summary_lbl = ttk.Label(self._ui_summary)

        # Layout
        self._ui_open.grid(column=0, row=0, sticky=tk.W)
        self._ui_save_as.grid(column=2, row=0, sticky=tk.W)
        # self._ui_buttons.columnconfigure(index=3, weight=1)

        self._ui_file_lbl.grid(column=0, row=0, sticky=tk.E)
        self._ui_file.grid(column=1, row=0, sticky=tk.EW)
        self._ui_file_btn.grid(column=2, row=0, sticky=tk.W)
        self._ui_buttons.grid(column=1, row=1, sticky=tk.EW, columnspan=2)
        self._ui_tab.columnconfigure(index=1, weight=1)

        self._ui_settings.grid(column=0, row=0, sticky=tk.NW)
        self._ui_summary_lbl.grid(column=1, row=0, sticky=tk.EW)
        self._ui_summary.columnconfigure(index=1, weight=1)

        self._ui_tab.grid(column=0, row=0, sticky=tk.EW)
        self._ui_summary.grid(column=0, row=1, sticky=tk.EW)
        self._ui_frm.columnconfigure(index=0, weight=1)

        self._ui_frm.grid(column=0, row=0, sticky=tk.EW)
        self.columnconfigure(index=0, weight=1)

        # Padding
        self._model.grid_slaves_configure(self._ui_tab, padx=1, pady=1)
        self._ui_tab.configure(padding=0)
        self._model.grid_slaves_configure(self._ui_buttons, padx=1, pady=1)
        self._ui_buttons.configure(padding=0)
        self._model.grid_slaves_configure(self._ui_summary, padx=1, pady=1)
        self._model.grid_slaves_configure(self._ui_frm, padx=5, pady=(0, 5))
        self._ui_frm.configure(padding=0)

        # Texts & icons
        self._ui_frm_lbl.configure(text='Configuration',
                                   image=self._model.icons.cfg_stage,
                                   compound=tk.LEFT)
        self._ui_file_lbl[TEXT] = 'Existing file:'
        self._ui_file_tip.text = 'Use empty string to load default configuration'
        self._ui_file_btn.configure(text='...', width=3)
        self._ui_open.configure(text='Open',
                                image=self._model.icons.open,
                                compound=tk.LEFT)
        self._ui_save_as.configure(text='Save as',
                                   image=self._model.icons.save_as,
                                   compound=tk.LEFT)
        self._ui_settings[IMAGE] = self._model.icons.settings
        self._ui_settings_tip.text = 'Edit common settings'

        # Hooks
        self._ui_file.bind(
            '<FocusOut>', lambda _e: self._on_file_leave())
        self._ui_file_btn[COMMAND] = self._on_get_file
        self._ui_open[COMMAND] = self._on_open
        self._ui_save_as[COMMAND] = self._on_save_as
        self._ui_settings[COMMAND] = self._on_settings
        self._ui_summary.bind(
            '<Configure>', lambda _e: self._ui_summary_lbl.configure(
                wraplength=self._ui_summary_lbl.winfo_width()))

        # Data
        self._var_file = tk.StringVar()
        self._ui_file[TEXTVARIABLE] = self._var_file
        self._var_settings = tk.IntVar()
        self._ui_settings[VARIABLE] = self._var_settings
        self._var_summary = tk.StringVar()
        self._ui_summary_lbl[TEXTVARIABLE] = self._var_summary

        self._update_thread: Optional[Thread] = None
        self._update_thread_err: Optional[str] = None

        self._cli_overrides: bool = True  # Resets after first update

        self._settings_dlg: Optional[ConfigCfgDialog] = None
        self._settings_apply: bool = False
        self._update_from_settings: bool = False

        self.stage_ui_init()

    def stage_type(self):
        return self._stage_type

    def stage_type_next(self):
        return self._stage_type_next

    def stage_is_active(self) -> bool:
        return True

    def stage_invalidate(self):
        if not self._update_from_settings:
            self._model.cfg = None

        self._model.stage_enabled(self._stage_type, False)

        self.stage_ui_init()
        self._model.stage_invalidate(self._stage_type_next)

    def stage_start_update(self):
        self.stage_invalidate()
        self._model.stages_ui_updating(True)
        self.winfo_toplevel().configure(cursor='watch')

        def _update_thread_fn():
            try:
                self._update_task()
            except BaseException as ex:
                self._update_thread_err = str(ex)
                tb.print_exception(None, ex, ex.__traceback__)

            def _update_done():
                self._update_thread = None
                self.winfo_toplevel().configure(cursor='')

                self._model.stages_ui_updating(False)

                if self._update_thread_err is not None:
                    err = self._update_thread_err
                    self._update_thread_err = None

                    detail = self._model.cfg_file
                    if self._model.cfg_file is None:
                        detail = '<default>'
                    messagebox.showerror(
                        title='File Error',
                        message=f'The configuration cannot be loaded:\n{err}',
                        detail=detail)

                    self._settings_apply = False
                    self._update_from_settings = False

                    self.stage_invalidate()
                    return

                if self._model.stage_is_active(self._stage_type_next):

                    # self._show_previews()

                    self._model.stage_ui_init(self._stage_type_next)
                    # Automatically start CSV parsing only if configuration file
                    # is not the default, CSV file is set, but image stack not
                    if (not self._update_from_settings
                            and self._model.cfg_file is not None
                            and self._model.cfg.csv_file is not None
                            and self._model.cfg.img_stack is None):
                        self._model.stage_start_update(self._stage_type_next)

                self._cli_overrides = False
                self._settings_apply = False
                self._update_from_settings = False

            self._model.invoke_on_gui_thread_async(_update_done)

        self._update_thread = Thread(target=_update_thread_fn, daemon=True)
        self._update_thread.start()

    def stage_ui_init(self):
        # CLI option takes precedence
        if self._cli_overrides:
            if self._model.cli_cfg_file is not None:
                self._model.cfg_file = self._model.cli_cfg_file

        if self._model.cfg_file is not None:
            self._var_file.set(str(self._model.cfg_file))

        self._ui_update_done()

    def stage_ui_updating(self, updating: bool):
        if updating:
            self._ui_update_start()
        else:
            self._ui_update_done()

    def _ui_update_start(self):
        self._ui_file.configure(state=tk.DISABLED)
        self._ui_file_btn.configure(state=tk.DISABLED)
        self._ui_open.configure(state=tk.DISABLED)
        self._ui_save_as.configure(state=tk.DISABLED)
        self._ui_settings.configure(state=tk.DISABLED)
        if self._update_from_settings:
            settings_dlg = self._settings_dlg
            if settings_dlg is not None:
                settings_dlg.enable(False)

    def _ui_update_done(self):
        self._ui_update_start()

        self._model.stage_enabled(self._stage_type, True)

        self._ui_file.configure(state=tk.NORMAL)
        self._ui_file_btn.configure(state=tk.NORMAL)
        self._ui_open.configure(state=tk.NORMAL)

        if self._model.cfg is not None:
            self._ui_save_as.configure(state=tk.NORMAL)
            self._ui_settings.configure(state=tk.NORMAL)
            if self._update_from_settings:
                settings_dlg = self._settings_dlg
                if settings_dlg is not None:
                    settings_dlg.enable(True)
            if self._model.cfg_file:
                self._var_summary.set(f'Loaded configuration from'
                                      f' {self._model.cfg_file.name}')
            else:
                self._var_summary.set('Loaded default configuration')
        else:
            self._var_summary.set('No configuration loaded yet')

    def _on_file_leave(self):
        file_name = self._var_file.get()
        self._model.cfg_file = Path(file_name) if file_name else None

    def _on_get_file(self):
        initial_dir = None
        initial_file = None
        if self._model.cfg_file is not None:
            initial_dir = self._model.cfg_file.parent
            initial_file = self._model.cfg_file.name

        file_name = filedialog.askopenfilename(
            parent=self,
            title='Select configuration file to open...',
            filetypes=(('JSON files', '*.json'), ('All files', '*')),
            initialdir=initial_dir,
            initialfile=initial_file)
        if file_name:
            self._var_file.set(file_name)
            self._model.cfg_file = Path(file_name)

    def _on_open(self):
        if self._model.cfg_file is not None:
            if not self._model.cfg_file.exists():
                messagebox.showerror(
                    title='File Error',
                    message='The configuration file does not exist.\n'
                            'Delete the path to load default configuration.',
                    detail=self._model.cfg_file)
                return

        self.stage_start_update()

    def _on_save_as(self):
        initial_dir = None
        initial_file = None
        if self._model.cfg_file is not None:
            path = self._model.cfg_file
            initial_dir = path.parent
            initial_file = path.name

        file_name = filedialog.asksaveasfilename(
            parent=self,
            title='Save configuration as...',
            filetypes=(('JSON files', '*.json'), ('All files', '*')),
            initialdir=initial_dir,
            initialfile=initial_file)
        if file_name:
            self._var_file.set(file_name)
            self._model.cfg_file = Path(file_name)
            self.save(self._model.cfg_file)

    def save(self, file_name: Path):
        try:
            cfg_dump = self._model.cfg.to_json()
            with open(file_name, 'wt', encoding='utf-8') as file:
                file.write(cfg_dump)
        except BaseException as ex:
            messagebox.showerror(
                title='File Error',
                message=f'The configuration cannot be saved:\n{str(ex)}',
                detail=file_name)

    def _on_settings(self):
        cfg_dump_old = self._model.cfg.to_json()

        def _process_cb(apply: bool):
            self._settings_apply = apply
            self._update_from_settings = True
            nonlocal cfg_dump_old
            cfg_dump_new = self._model.cfg.to_json()
            if (cfg_dump_old != cfg_dump_new
                    or not self._model.stage_is_active(self._stage_type_next)):
                cfg_dump_old = cfg_dump_new
                self.stage_start_update()
            else:
                # self._show_previews()
                self._settings_apply = False
                self._update_from_settings = False

        self._settings_dlg = ConfigCfgDialog(self, self._model, process_cb=_process_cb)
        self._settings_dlg.show_modal()
        self._settings_dlg = None
        self._settings_apply = False
        self._update_from_settings = False
        self._var_settings.set(0)

    # Executed on thread
    def _update_task(self):
        if not self._update_from_settings:
            self._model.invoke_on_gui_thread_async(
                self._var_summary.set, 'Loading configuration...')

            if self._model.cfg_file is not None:
                with open(self._model.cfg_file, 'rt', encoding='utf-8') as file:
                    cfg_dump = file.read()
            else:
                cfg_dump = pkgutil.get_data(
                    smlfm.__name__, 'data/default-config.json').decode()

            self._model.cfg = smlfm.Config.from_json(cfg_dump)

        # CLI option takes precedence
        if self._cli_overrides or self._model.cfg.img_stack is None:
            if self._model.cli_img_stack is not None:
                self._model.cfg.img_stack = self._model.cli_img_stack
        if self._cli_overrides or self._model.cfg.csv_file is None:
            if self._model.cli_csv_file is not None:
                self._model.cfg.csv_file = self._model.cli_csv_file

        self._model.invoke_on_gui_thread_async(
            self._var_summary.set, 'Detecting Fiji...')
        has_fiji = self._model.fiji_app.detect_fiji(
            self._model.cfg.fiji_dir, self._model.cfg.fiji_jvm_opts)
        if has_fiji is None:
            print('WARNING: Restart the application to apply new settings')
            self._model.invoke_on_gui_thread_async(
                messagebox.showwarning,
                kwargs={
                    'title': 'Warning',
                    'message': 'Restart the application to apply new settings.',
                    'detail': 'Java can be started only once.',
                }
            )



import time
import tkinter as tk
import traceback as tb
from idlelib.tooltip import Hovertip
from pathlib import Path
from threading import Thread
from tkinter import filedialog, messagebox, ttk
from typing import Optional

import numpy as np
from matplotlib.figure import Figure

import smlfm
import smlfm.graphs

from .app_model import AppModel, IStage
from .consts import COMMAND, IMAGE, READONLY, TEXT, TEXTVARIABLE, VARIABLE, GraphType, StageType
from .figure_window import FigureWindow


# pylint: disable=too-many-ancestors,too-many-instance-attributes
class CsvFrame(ttk.Frame, IStage):

    def __init__(self, master, model: AppModel, *args, **kwargs):
        super().__init__(master, *args, **kwargs)

        self._stage_type = StageType.CSV
        self._stage_type_next = StageType.OPTICS

        self._model = model

        cfg_default = smlfm.Config()

        # Controls
        self._ui_frm_lbl = ttk.Label(self)
        self._ui_frm = ttk.LabelFrame(self, labelwidget=self._ui_frm_lbl)

        self._ui_tabs = ttk.Notebook(self._ui_frm)
        self._ui_tabs_switched = False  # Might be switched once to _ui_tab2

        self._ui_tab1 = ttk.Frame(self._ui_tabs)
        self._ui_tabs.add(self._ui_tab1)
        self._ui_file_loc_lbl = ttk.Label(self._ui_tab1)
        self._ui_file_loc = ttk.Entry(self._ui_tab1)
        # noinspection PyProtectedMember
        Hovertip(self._ui_file_loc, text=cfg_default._csv_file_doc)
        self._ui_file_loc_btn = ttk.Button(self._ui_tab1)
        self._ui_fmt_loc_lbl = ttk.Label(self._ui_tab1)
        self._ui_fmt_loc = ttk.Combobox(self._ui_tab1, state=READONLY)
        self._ui_open = ttk.Button(self._ui_tab1)

        self._ui_tab2 = ttk.Frame(self._ui_tabs)
        self._ui_tabs.add(self._ui_tab2)
        self._ui_file_img_lbl = ttk.Label(self._ui_tab2)
        self._ui_file_img = ttk.Entry(self._ui_tab2)
        # noinspection PyProtectedMember
        Hovertip(self._ui_file_img, text=cfg_default._img_stack_doc)
        self._ui_file_img_btn = ttk.Button(self._ui_tab2)
        self._ui_peakfit = ttk.Button(self._ui_tab2)
        self._ui_no_fiji_lbl = ttk.Label(self._ui_tab2)

        self._ui_summary = ttk.Frame(self._ui_frm)
        self._ui_summary_lbl = ttk.Label(self._ui_summary)
        self._ui_preview = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_preview_tip = Hovertip(self._ui_preview, text='')

        # Layout
        self._ui_file_loc_lbl.grid(column=0, row=0, sticky=tk.E)
        self._ui_file_loc.grid(column=1, row=0, sticky=tk.EW)
        self._ui_file_loc_btn.grid(column=2, row=0, sticky=tk.W)
        self._ui_fmt_loc_lbl.grid(column=0, row=1, sticky=tk.E)
        self._ui_fmt_loc.grid(column=1, row=1, sticky=tk.EW, columnspan=2)
        self._ui_open.grid(column=1, row=2, sticky=tk.W)
        self._ui_tab1.columnconfigure(index=1, weight=1)

        self._ui_file_img_lbl.grid(column=0, row=0, sticky=tk.E)
        self._ui_file_img.grid(column=1, row=0, sticky=tk.EW, columnspan=2)
        self._ui_file_img_btn.grid(column=3, row=0, sticky=tk.W)
        self._ui_peakfit.grid(column=1, row=1, sticky=tk.W)
        self._ui_no_fiji_lbl.grid(column=2, row=1, sticky=tk.EW, columnspan=2)
        self._ui_tab2.columnconfigure(index=2, weight=1)

        self._ui_tabs.columnconfigure(index=0, weight=1)

        self._ui_summary_lbl.grid(column=0, row=0, sticky=tk.EW)
        self._ui_preview.grid(column=1, row=0, sticky=tk.NW)
        self._ui_summary.columnconfigure(index=0, weight=1)

        self._ui_tabs.grid(column=0, row=0, sticky=tk.EW)
        self._ui_summary.grid(column=0, row=1, sticky=tk.EW)
        self._ui_frm.columnconfigure(index=0, weight=1)

        self._ui_frm.grid(column=0, row=0, sticky=tk.EW)
        self.columnconfigure(index=0, weight=1)

        # Padding
        self._model.grid_slaves_configure(self._ui_tab1, padx=1, pady=1)
        self._ui_tab1.configure(padding=5)
        self._model.grid_slaves_configure(self._ui_tab2, padx=1, pady=1)
        self._ui_tab2.configure(padding=5)
        self._ui_tabs.configure(padding=0)
        self._model.grid_slaves_configure(self._ui_summary, padx=1, pady=1)
        self._model.grid_slaves_configure(self._ui_frm, padx=5, pady=(0, 5))
        self._ui_frm.configure(padding=0)

        # Texts & icons
        self._ui_frm_lbl.configure(text='Localisations',
                                   image=self._model.icons.csv_stage,
                                   compound=tk.LEFT)
        self._ui_tabs.tab(self._ui_tab1,
                          text='I have CSV file',
                          image=self._model.icons.csv_data,
                          compound=tk.LEFT)
        self._ui_file_loc_lbl[TEXT] = 'Existing file:'
        self._ui_file_loc_btn.configure(text='...', width=3)
        self._ui_fmt_loc_lbl[TEXT] = 'File format:'
        self._ui_open.configure(text='Open',
                                image=self._model.icons.open,
                                compound=tk.LEFT)
        self._ui_tabs.tab(self._ui_tab2,
                          text='I have image stack',
                          image=self._model.icons.csv_stack,
                          compound=tk.LEFT)
        self._ui_file_img_lbl[TEXT] = 'Image stack:'
        self._ui_file_img_btn.configure(text='...', width=3)
        self._ui_peakfit.configure(text='Run PeakFit in Fiji',
                                   image=self._model.icons.csv_peakfit,
                                   compound=tk.LEFT)
        self._ui_no_fiji_lbl.configure(text='Fiji not detected')
        self._ui_preview[IMAGE] = self._model.icons.preview
        self._ui_preview_tip.text = 'Preview raw localisations'

        # Hooks
        self._ui_file_loc.bind(
            '<FocusOut>', lambda _e: self._on_file_loc_leave())
        self._ui_file_loc_btn[COMMAND] = self._on_get_file_loc
        self._ui_fmt_loc.bind(
            '<FocusOut>', lambda _e: self._on_fmt_loc_leave())
        self._ui_open[COMMAND] = self._on_open
        self._ui_file_img.bind(
            '<FocusOut>', lambda _e: self._on_file_img_leave())
        self._ui_file_img_btn[COMMAND] = self._on_get_file_img
        self._ui_peakfit[COMMAND] = self._on_run_peakfit
        self._ui_summary.bind(
            '<Configure>', lambda _e: self._ui_summary_lbl.configure(
                wraplength=self._ui_summary_lbl.winfo_width()))
        self._ui_preview[COMMAND] = self._on_preview

        # Data
        self._var_file_loc = tk.StringVar()
        self._ui_file_loc[TEXTVARIABLE] = self._var_file_loc
        self._var_fmt_loc = tk.StringVar()
        self._ui_fmt_loc.configure(
            textvariable=self._var_fmt_loc,
            values=[f.name for f in smlfm.LocalisationFile.Format])
        self._var_file_img = tk.StringVar()
        self._ui_file_img[TEXTVARIABLE] = self._var_file_img
        self._var_summary = tk.StringVar()
        self._ui_summary_lbl[TEXTVARIABLE] = self._var_summary
        self._var_preview = tk.IntVar()
        self._ui_preview[VARIABLE] = self._var_preview

        self._update_thread: Optional[Thread] = None
        self._update_thread_err: Optional[str] = None

        self.stage_ui_init()

    def stage_type(self):
        return self._stage_type

    def stage_type_next(self):
        return self._stage_type_next

    def stage_is_active(self) -> bool:
        return self._model.cfg is not None

    def stage_invalidate(self):
        self._model.csv = None

        self._model.destroy_graph(GraphType.CSV_RAW)

        self._model.stage_enabled(self._stage_type, False)

        self.stage_ui_init()
        self._model.stage_invalidate(self._stage_type_next)

    def stage_start_update(self):
        self.stage_invalidate()
        self._model.stages_ui_updating(True)
        self.winfo_toplevel().configure(cursor='watch')

        def _update_thread_fn():
            try:
                self._update_task()
            except BaseException as ex:
                self._update_thread_err = str(ex)
                tb.print_exception(None, ex, ex.__traceback__)

            def _update_done():
                self._update_thread = None
                self.winfo_toplevel().configure(cursor='')

                self._model.stages_ui_updating(False)

                if self._update_thread_err is not None:
                    err = self._update_thread_err
                    self._update_thread_err = None

                    messagebox.showerror(
                        title='File Error',
                        message=f'The localisations cannot be loaded:\n{str(err)}',
                        detail=self._model.cfg.csv_file)

                    self.stage_invalidate()
                    return

                if self._model.stage_is_active(self._stage_type_next):

                    self._show_previews(force_update=True)

                    self._model.stage_ui_init(self._stage_type_next)
                    self._model.stage_start_update(self._stage_type_next)

            self._model.invoke_on_gui_thread_async(_update_done)

        self._update_thread = Thread(target=_update_thread_fn, daemon=True)
        self._update_thread.start()

    def stage_ui_init(self):
        if self._model.cfg is not None:
            if self._model.cfg.csv_file is not None:
                self._var_file_loc.set(str(self._model.cfg.csv_file))

            self._var_fmt_loc.set(self._model.cfg.csv_format.name)

            if self._model.cfg.img_stack is not None:
                self._var_file_img.set(str(self._model.cfg.img_stack))
                if self._model.fiji_app.has_fiji and not self._ui_tabs_switched:
                    self._ui_tabs.select(self._ui_tab2)
            self._ui_tabs_switched = True

        if self._model.fiji_app.has_fiji:
            self._ui_no_fiji_lbl.grid_forget()

        self._ui_update_done()

    def stage_ui_updating(self, updating: bool):
        if updating:
            self._ui_update_start()
        else:
            self._ui_update_done()

    def _ui_update_start(self):
        self._ui_file_loc.configure(state=tk.DISABLED)
        self._ui_file_loc_btn.configure(state=tk.DISABLED)
        self._ui_fmt_loc.configure(state=tk.DISABLED)
        self._ui_open.configure(state=tk.DISABLED)

        self._ui_file_img.configure(state=tk.DISABLED)
        self._ui_file_img_btn.configure(state=tk.DISABLED)
        self._ui_peakfit.configure(state=tk.DISABLED)

        self._ui_preview.configure(state=tk.DISABLED)

    def _ui_update_done(self):
        self._ui_update_start()

        if self.stage_is_active():
            self._model.stage_enabled(self._stage_type, True)

            self._ui_file_loc.configure(state=tk.NORMAL)
            self._ui_file_loc_btn.configure(state=tk.NORMAL)
            self._ui_fmt_loc.configure(state=READONLY)
            self._ui_open.configure(state=tk.NORMAL)

            if self._model.fiji_app.has_fiji:
                self._ui_file_img.configure(state=tk.NORMAL)
                self._ui_file_img_btn.configure(state=tk.NORMAL)
                self._ui_peakfit.configure(state=tk.NORMAL)

            if self._model.stage_is_active(self._stage_type_next):
                self._ui_preview.configure(state=tk.NORMAL)
                locs = self._model.csv.data.shape[0]
                frames = np.unique(self._model.csv.data[:, 0]).shape[0]
                self._var_summary.set(
                    f'Loaded {locs} localisations from {frames} unique frames')
            else:
                self._var_summary.set('No localisations loaded')
        else:
            self._var_summary.set('No localisations loaded yet')

    def _on_file_loc_leave(self):
        file_name = self._var_file_loc.get()
        self._model.cfg.csv_file = Path(file_name) if file_name else None

    def _on_get_file_loc(self):
        initial_dir = None
        initial_file = None
        if self._model.cfg.csv_file is not None:
            initial_dir = self._model.cfg.csv_file.parent
            initial_file = self._model.cfg.csv_file.name

        file_name = filedialog.askopenfilename(
            parent=self,
            title='Select CSV file to open...',
            filetypes=(('CSV files', '*.csv *.xls'), ('All files', '*')),
            initialdir=initial_dir,
            initialfile=initial_file)
        if file_name:
            self._var_file_loc.set(file_name)
            self._model.cfg.csv_file = Path(file_name)

    def _on_fmt_loc_leave(self):
        fmt = self._var_fmt_loc.get()
        self._model.cfg.csv_format = smlfm.LocalisationFile.Format[fmt]

    def _on_file_img_leave(self):
        file_name = self._var_file_img.get()
        self._model.cfg.img_stack = Path(file_name) if file_name else None

    def _on_get_file_img(self):
        initial_dir = None
        initial_file = None
        if self._model.cfg.img_stack is not None:
            initial_dir = self._model.cfg.img_stack.parent
            initial_file = self._model.cfg.img_stack.name
        else:
            if self._model.cfg.csv_file is not None:
                initial_dir = self._model.cfg.csv_file.parent

        file_name = filedialog.askopenfilename(
            parent=self,
            title='Select image stack file to open...',
            filetypes=(('TIFF files', '*.tiff *.tif'), ('All files', '*')),
            initialdir=initial_dir,
            initialfile=initial_file)
        if file_name:
            self._var_file_img.set(file_name)
            self._model.cfg.img_stack = Path(file_name)

    def _on_open(self):
        if self._model.cfg.csv_file is None:
            messagebox.showerror(
                title='File Error',
                message='The CSV file path cannot be empty.')
            return

        if not self._model.cfg.csv_file.exists():
            messagebox.showerror(
                title='File Error',
                message='The CSV file does not exist:',
                detail=self._model.cfg.csv_file)
            return

        self.stage_start_update()

    def _show_previews(self, force_update: bool = False):
        if not self._model.stage_is_active(self._stage_type_next):
            return

        show = self._model.cfg.show_graphs and self._model.cfg.show_all_debug_graphs

        force_show = show  # or self._settings_apply
        force_update = force_update and self._model.graph_exists(GraphType.CSV_RAW)
        if force_show:
            self._var_preview.set(1)
        if force_show or force_update:
            self._on_preview(force_update)

    def _on_preview(self, force_update: bool = False):

        def draw(f: Figure, set_size: bool = True) -> Figure:
            return smlfm.graphs.draw_locs_csv(
                f,
                self._model.csv.data[:, 1:3],
                set_default_size=set_size)

        wnd = self._model.graphs[GraphType.CSV_RAW]
        if wnd is None:
            fig = draw(Figure())
            wnd = FigureWindow(fig, master=self, title='Raw localisations')
            wnd.bind('<Map>', func=lambda _evt: self._var_preview.set(1))
            wnd.bind('<Unmap>', func=lambda _evt: self._var_preview.set(0))
            self._model.graphs[GraphType.CSV_RAW] = wnd
        else:
            if force_update:
                draw(wnd.figure, set_size=False)
                wnd.refresh()

        if self._var_preview.get():
            wnd.deiconify()
        else:
            wnd.withdraw()

    def _on_run_peakfit(self):
        if self._model.cfg.img_stack is None:
            messagebox.showerror(
                title='File Error',
                message='The path to image stack file cannot be empty.')
            return

        if not self._model.cfg.img_stack.exists():
            messagebox.showerror(
                title='File Error',
                message='The image stack file does not exist:',
                detail=self._model.cfg.img_stack)
            return

        if self._model.cfg.csv_file is None:
            file_name = str(self._model.cfg.img_stack) + '.csv'
            self._var_file_loc.set(file_name)
            self._model.cfg.csv_file = Path(file_name)

        self._model.cfg.csv_format = smlfm.LocalisationFile.Format.PEAKFIT
        self._var_fmt_loc.set(self._model.cfg.csv_format.name)

        self._start_peakfit()

    def _start_peakfit(self):
        self.stage_invalidate()
        self._model.stages_ui_updating(True)
        self.winfo_toplevel().configure(cursor='watch')
        self._var_summary.set('Searching localisations...')

        def _update_thread_fn():
            try:
                self._model.fiji_app.run_peakfit(
                    self._model.cfg.img_stack, self._model.cfg.csv_file)
            except BaseException as ex:
                self._update_thread_err = str(ex)
                tb.print_exception(None, ex, ex.__traceback__)

            def _update_done():
                self._update_thread = None
                self.winfo_toplevel().configure(cursor='')

                self._model.stages_ui_updating(False)

                if self._update_thread_err is not None:
                    err = self._update_thread_err
                    self._update_thread_err = None

                    messagebox.showerror(
                        title='Fiji Error',
                        message=f'The PeakFit plugin failed with:\n{str(err)}',
                        detail=self._model.cfg.img_stack)

                    self.stage_invalidate()
                    return

                if not self._model.cfg.csv_file.exists():
                    messagebox.showerror(
                        title='File Error',
                        message='The CSV file does not exist:',
                        detail=self._model.cfg.csv_file)

                    self.stage_invalidate()
                    return

                self.stage_start_update()

            self._model.invoke_on_gui_thread_async(_update_done)

        self._update_thread = Thread(target=_update_thread_fn, daemon=True)
        self._update_thread.start()

    # Executed on thread
    def _update_task(self):
        self._model.invoke_on_gui_thread_async(
            self._var_summary.set, 'Loading localisations...')
        tic = time.time()

        self._model.csv = smlfm.LocalisationFile(
            self._model.cfg.csv_file, self._model.cfg.csv_format)
        self._model.csv.read()

        if self._model.cfg.log_timing:
            print(f'Loading {repr(self._model.cfg.csv_file.name)} took'
                  f' {1e3 * (time.time() - tic):.3f} ms')



import dataclasses
import multiprocessing as mp
import time
import tkinter as tk
import traceback as tb
from datetime import datetime
from idlelib.tooltip import Hovertip
from pathlib import Path
from threading import Thread
from tkinter import filedialog, messagebox, ttk
from typing import Optional

import numpy as np
import numpy.typing as npt
from matplotlib.figure import Figure

import smlfm
import smlfm.graphs

from .app_model import AppModel, IStage
from .consts import COMMAND, IMAGE, TEXT, TEXTVARIABLE, VARIABLE, GraphType, StageType
from .figure_window import FigureWindow
from .fit_cfg_dialog import FitCfgDialog


# pylint: disable=too-many-ancestors,too-many-instance-attributes
class FitFrame(ttk.Frame, IStage):

    # pylint: disable=too-many-statements
    def __init__(self, master, model: AppModel, *args, **kwargs):
        super().__init__(master, *args, **kwargs)

        self._stage_type = StageType.FIT
        self._stage_type_next = None

        self._model = model

        # Controls
        self._ui_frm_lbl = ttk.Label(self)
        self._ui_frm = ttk.LabelFrame(self, labelwidget=self._ui_frm_lbl)

        self._ui_tab = ttk.Frame(self._ui_frm)
        self._ui_autosave_dir_chb = ttk.Checkbutton(self._ui_tab)
        self._ui_autosave_dir_chb_tip = Hovertip(self._ui_autosave_dir_chb, text='')
        self._ui_save_dir = ttk.Entry(self._ui_tab)
        self._ui_save_dir_tip = Hovertip(self._ui_save_dir, text='')
        self._ui_save_dir_btn = ttk.Button(self._ui_tab)
        self._ui_save_as = ttk.Button(self._ui_tab)

        self._ui_summary = ttk.Frame(self._ui_frm)
        self._ui_settings = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_settings_tip = Hovertip(self._ui_settings, text='')
        self._ui_start = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_start_tip = Hovertip(self._ui_start, text='')
        self._ui_summary_lbl = ttk.Label(self._ui_summary)
        self._ui_preview_occ = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_preview_occ_tip = Hovertip(self._ui_preview_occ, text='')
        self._ui_preview_hist = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_preview_hist_tip = Hovertip(self._ui_preview_hist, text='')
        self._ui_preview_3d = ttk.Checkbutton(self._ui_summary, style='Toolbutton')
        self._ui_preview_3d_tip = Hovertip(self._ui_preview_3d, text='')

        self._ui_progress = ttk.Progressbar(
            self._ui_frm, orient=tk.HORIZONTAL, mode='determinate')

        # Layout
        self._ui_autosave_dir_chb.grid(column=0, row=0, sticky=tk.E)
        self._ui_save_dir.grid(column=1, row=0, sticky=tk.EW)
        self._ui_save_dir_btn.grid(column=2, row=0, sticky=tk.W)
        self._ui_save_as.grid(column=1, row=1, sticky=tk.W)
        self._ui_tab.columnconfigure(index=1, weight=1)

        self._ui_settings.grid(column=0, row=0, sticky=tk.NW)
        self._ui_start.grid(column=1, row=0, sticky=tk.NW)
        self._ui_summary_lbl.grid(column=2, row=0, sticky=tk.EW)
        self._ui_preview_occ.grid(column=3, row=0, sticky=tk.NW)
        self._ui_preview_hist.grid(column=4, row=0, sticky=tk.NW)
        self._ui_preview_3d.grid(column=5, row=0, sticky=tk.NW)
        self._ui_summary.columnconfigure(index=2, weight=1)

        self._ui_tab.grid(column=0, row=0, sticky=tk.EW)
        self._ui_summary.grid(column=0, row=1, sticky=tk.EW)
        self._ui_progress.grid(column=0, row=2, sticky=tk.EW)
        self._ui_frm.columnconfigure(index=0, weight=1)

        self._ui_frm.grid(column=0, row=0, sticky=tk.EW)
        self.columnconfigure(index=0, weight=1)

        # Padding
        self._model.grid_slaves_configure(self._ui_tab, padx=1, pady=1)
        self._ui_tab.configure(padding=0)
        self._model.grid_slaves_configure(self._ui_summary, padx=1, pady=1)
        self._model.grid_slaves_configure(self._ui_frm, padx=5, pady=(0, 5))
        self._ui_frm.configure(padding=0)

        # Texts & icons
        self._ui_frm_lbl.configure(text='Fitting',
                                   image=self._model.icons.fit_stage,
                                   compound=tk.LEFT)
        self._ui_autosave_dir_chb[TEXT] = 'Autosave to:'
        self._ui_autosave_dir_chb_tip.text = (
            'Results are saved under this directory to a new sub-folder\n'
            'with a name containing a timestamp of execution.')
        self._ui_save_dir_tip.text = self._ui_autosave_dir_chb_tip.text
        self._ui_save_dir_btn.configure(text='...', width=3)
        self._ui_save_as.configure(text='Save as',
                                   image=self._model.icons.save_as,
                                   compound=tk.LEFT)
        self._ui_settings[IMAGE] = self._model.icons.settings
        self._ui_settings_tip.text = 'Edit parameters'
        self._ui_start[IMAGE] = self._model.icons.start
        self._ui_start_tip.text = 'Proceed with fitting'
        self._ui_preview_occ[IMAGE] = self._model.icons.fit_occurrences
        self._ui_preview_occ_tip.text = 'Preview occurrences'
        self._ui_preview_hist[IMAGE] = self._model.icons.fit_histogram
        self._ui_preview_hist_tip.text = 'Preview histogram'
        self._ui_preview_3d[IMAGE] = self._model.icons.fit_3d
        self._ui_preview_3d_tip.text = 'Final 3D localisations'
        self._ui_progress.configure()

        # Hooks
        self._ui_autosave_dir_chb[COMMAND] = self._on_autosave_dir
        self._ui_save_dir.bind(
            '<FocusOut>', lambda _e: self._on_save_dir_leave())
        self._ui_save_dir_btn[COMMAND] = self._on_get_save_dir
        self._ui_save_as[COMMAND] = self._on_save_as
        self._ui_settings[COMMAND] = self._on_settings
        self._ui_start[COMMAND] = self._on_start
        self._ui_summary.bind(
            '<Configure>', lambda _e: self._ui_summary_lbl.configure(
                wraplength=self._ui_summary_lbl.winfo_width()))
        self._ui_preview_occ[COMMAND] = self._on_preview_occ
        self._ui_preview_hist[COMMAND] = self._on_preview_hist
        self._ui_preview_3d[COMMAND] = self._on_preview_3d

        # Data
        self._var_autosave_dir_chb = tk.IntVar()
        self._ui_autosave_dir_chb[VARIABLE] = self._var_autosave_dir_chb
        self._var_save_dir = tk.StringVar()
        self._ui_save_dir[TEXTVARIABLE] = self._var_save_dir
        self._var_settings = tk.IntVar()
        self._ui_settings[VARIABLE] = self._var_settings
        self._var_start = tk.IntVar()
        self._ui_start[VARIABLE] = self._var_start
        self._var_summary = tk.StringVar()
        self._ui_summary_lbl[TEXTVARIABLE] = self._var_summary
        self._var_preview_occ = tk.IntVar()
        self._ui_preview_occ[VARIABLE] = self._var_preview_occ
        self._var_preview_hist = tk.IntVar()
        self._ui_preview_hist[VARIABLE] = self._var_preview_hist
        self._var_preview_3d = tk.IntVar()
        self._ui_preview_3d[VARIABLE] = self._var_preview_3d
        self._var_progress = tk.IntVar()
        self._ui_progress[VARIABLE] = self._var_progress

        self._update_thread: Optional[Thread] = None
        self._update_thread_err: Optional[str] = None
        self._update_thread_abort: Optional[mp.Event] = None

        self._settings_dlg: Optional[FitCfgDialog] = None
        self._settings_apply: bool = False
        self._update_from_settings: bool = False

        # =0 ~ disabled, >0 ~ remaining flashes, None ~ infinite
        self._flash_counter = 0
        self._flash_id = None

        # A cache with locs_3d filtered by max_lateral_err and min_view_count
        self._keep_idx: Optional[npt.NDArray[float]] = None

        self._saved_path: Optional[Path] = None

        self.stage_ui_init()

    def stage_type(self):
        return self._stage_type

    def stage_type_next(self):
        return self._stage_type_next

    def stage_is_active(self) -> bool:
        lfl = self._model.lfl
        return lfl is not None and lfl.correction is not None

    def stage_invalidate(self):
        self._flash_stop()

        self._model.locs_3d = None
        self._keep_idx = None
        self._saved_path = None

        if not self._update_from_settings:
            for gt in [GraphType.LOCS_3D, GraphType.HISTOGRAM, GraphType.OCCURRENCES]:
                self._model.destroy_graph(gt)

        self._model.stage_enabled(self._stage_type, False)

        self.stage_ui_init()
        self._model.stage_invalidate(self._stage_type_next)

    def stage_start_update(self):
        self.stage_invalidate()
        self._model.stages_ui_updating(True)
        self.winfo_toplevel().configure(cursor='watch')

        self._model.save_timestamp = datetime.now()

        self._ui_start[IMAGE] = self._model.icons.cancel
        self._ui_start_tip.text = 'Abort processing'
        self._ui_start.configure(state=tk.NORMAL)

        frame_min = (self._model.cfg.fit_params_full.frame_min
                     if self._model.cfg.fit_params_full.frame_min > 0
                     else self._model.lfl.min_frame)
        frame_max = (self._model.cfg.fit_params_full.frame_max
                     if self._model.cfg.fit_params_full.frame_max > 0
                     else self._model.lfl.max_frame)
        self._ui_progress.configure(maximum=frame_max - frame_min + 1)
        self._var_progress.set(0)

        def _update_thread_fn():
            try:
                self._update_task()
            except BaseException as ex:
                self._update_thread_err = str(ex)
                tb.print_exception(None, ex, ex.__traceback__)

            def _update_done():
                self._ui_start[IMAGE] = self._model.icons.start
                self._ui_start_tip.text = 'Proceed with fitting'
                self._ui_start.configure(state=tk.DISABLED)

                self._update_thread = None
                self.winfo_toplevel().configure(cursor='')

                self._model.stages_ui_updating(False)

                # Reset the flag after the stage update prints it was aborted
                self._update_thread_abort = None

                if self._update_thread_err is not None:
                    err = self._update_thread_err
                    self._update_thread_err = None

                    if self._model.locs_3d is None:
                        messagebox.showerror(
                            title='Fitting Error',
                            message=f'The fitting failed:\n{err}')
                    else:
                        messagebox.showerror(
                            title='File Error',
                            message=f'The results cannot be saved:\n{err}')

                    self._settings_apply = False
                    self._update_from_settings = False

                    self.stage_invalidate()
                    return

                # if self._model.stage_is_active(self._stage_type_next):
                if self._model.locs_3d is not None:

                    self._show_previews(force_update=True)

                    self._model.stage_ui_init(self._stage_type_next)
                    # Next phase updates automatically or the user continues manually
                    if self._update_from_settings:
                        self._model.stage_ui_flash(self._stage_type_next)
                    else:
                        if self._model.cfg.confirm_mla_alignment:
                            self._model.stage_ui_flash(self._stage_type_next)
                        else:
                            self._model.stage_start_update(self._stage_type_next)

                self._settings_apply = False
                self._update_from_settings = False

            self._model.invoke_on_gui_thread_async(_update_done)

        self._update_thread_abort = mp.Event()
        self._update_thread = Thread(target=_update_thread_fn, daemon=True)
        self._update_thread.start()

    def stage_ui_init(self):
        if self._model.cfg is not None:
            if self._model.cfg.save_dir is None:
                self._var_autosave_dir_chb.set(0)
                self._var_save_dir.set('')
            else:
                self._var_autosave_dir_chb.set(1)
                self._var_save_dir.set(str(self._model.cfg.save_dir))

        self._ui_update_done()

    def stage_ui_flash(self):
        self._flash_start()

    def stage_ui_updating(self, updating: bool):
        if updating:
            self._ui_update_start()
        else:
            self._ui_update_done()

    def _ui_update_start(self):
        self._ui_autosave_dir_chb.configure(state=tk.DISABLED)
        self._ui_save_dir.configure(state=tk.DISABLED)
        self._ui_save_dir_btn.configure(state=tk.DISABLED)
        self._ui_save_as.configure(state=tk.DISABLED)
        self._ui_settings.configure(state=tk.DISABLED)
        if self._update_from_settings:
            settings_dlg = self._settings_dlg
            if settings_dlg is not None:
                settings_dlg.enable(False)
        self._ui_start.configure(state=tk.DISABLED)
        self._ui_preview_occ.configure(state=tk.DISABLED)
        self._ui_preview_hist.configure(state=tk.DISABLED)
        self._ui_preview_3d.configure(state=tk.DISABLED)

    # pylint: disable=too-many-branches
    def _ui_update_done(self):
        self._ui_update_start()

        if self.stage_is_active():
            self._model.stage_enabled(self._stage_type, True)

            self._ui_autosave_dir_chb.configure(state=tk.NORMAL)
            if self._var_autosave_dir_chb.get():
                self._ui_save_dir.configure(state=tk.NORMAL)
                self._ui_save_dir_btn.configure(state=tk.NORMAL)
            else:
                self._ui_save_dir.configure(state=tk.DISABLED)
                self._ui_save_dir_btn.configure(state=tk.DISABLED)
            self._ui_settings.configure(state=tk.NORMAL)
            if self._update_from_settings:
                settings_dlg = self._settings_dlg
                if settings_dlg is not None:
                    settings_dlg.enable(True)
            self._ui_start.configure(state=tk.NORMAL)

            # if self._model.stage_is_active(self._stage_type_next):
            if self._model.locs_3d is not None:
                if self._var_autosave_dir_chb.get():
                    self._ui_save_as.configure(state=tk.DISABLED)
                else:
                    self._ui_save_as.configure(state=tk.NORMAL)
                self._ui_preview_occ.configure(state=tk.NORMAL)
                self._ui_preview_hist.configure(state=tk.NORMAL)
                self._ui_preview_3d.configure(state=tk.NORMAL)

                points = self._model.locs_3d.shape[0]
                frames = np.unique(self._model.locs_3d[:, 7]).shape[0]
                views = int(np.sum(self._model.locs_3d[:, 5]))
                fit_msg = (
                    f'Fitting completed with {points} 3D localisations,'
                    f' used {frames} frames and {views} 2D localisations in total')
                if self._saved_path is not None:
                    self._var_summary.set(
                        f'{fit_msg}.\n'
                        f'Results saved to folder: "{self._saved_path.name}"')
                else:
                    self._var_summary.set(fit_msg)
            else:
                if (self._update_thread_abort is not None
                        and self._update_thread_abort.is_set()):
                    self._var_summary.set('Fitting aborted')
                else:
                    self._var_summary.set('No fitting done')
        else:
            self._var_summary.set('No fitting done yet')

        # Set progress bar full
        self._ui_progress.stop()
        self._ui_progress.configure(maximum=100)
        self._var_progress.set(100)

    def _flash_start(self):
        self._flash_stop()

        def _toggle_on():
            if self._flash_counter is None or self._flash_counter > 0:
                self._var_start.set(1)
                self._flash_id = self.after(500, _toggle_off)

        def _toggle_off():
            self._var_start.set(0)
            if self._flash_counter is None or self._flash_counter > 0:
                if self._flash_counter is not None:
                    self._flash_counter -= 1
                self._flash_id = self.after(500, _toggle_on)

        self._flash_counter = None
        _toggle_on()

    def _flash_stop(self):
        self._flash_counter = 0
        if self._flash_id is not None:
            self.after_cancel(self._flash_id)
        self._var_start.set(0)

    def _on_autosave_dir(self):
        if self._var_autosave_dir_chb.get():
            dir_name = self._var_save_dir.get()
            if dir_name:
                self._model.cfg.save_dir = Path(dir_name)
            else:
                self._model.cfg.save_dir = None

            self._ui_save_dir.configure(state=tk.NORMAL)
            self._ui_save_dir_btn.configure(state=tk.NORMAL)
            if self._model.locs_3d is not None:
                self._ui_save_as.configure(state=tk.DISABLED)
        else:
            self._model.cfg.save_dir = None

            self._ui_save_dir.configure(state=tk.DISABLED)
            self._ui_save_dir_btn.configure(state=tk.DISABLED)
            if self._model.locs_3d is not None:
                self._ui_save_as.configure(state=tk.NORMAL)

    def _on_save_dir_leave(self):
        dir_name = self._var_save_dir.get()
        self._model.cfg.save_dir = Path(dir_name) if dir_name else None

    def _on_get_save_dir(self):
        initial_dir = None
        if self._model.cfg.save_dir is not None:
            initial_dir = self._model.cfg.save_dir

        dir_name = filedialog.askdirectory(
            parent=self,
            title='Select top-level folder for results...',
            initialdir=initial_dir)
        if dir_name:
            self._var_save_dir.set(dir_name)
            self._model.cfg.save_dir = Path(dir_name)

    def _on_save_as(self):
        initial_dir = None
        if self._model.cfg.save_dir is not None:
            initial_dir = self._model.cfg.save_dir

        dir_name = filedialog.askdirectory(
            parent=self,
            title='Select folder for results...',
            initialdir=initial_dir)
        if dir_name:
            bak = self._model.cfg.save_dir
            try:
                # Resolve to absolute path to not get a sub-folder with timestamp
                # Change save path temporarily to have it stored with results,
                # but enabled autosave afterward.
                self._model.cfg.save_dir = Path(dir_name).resolve()
                self._saved_path = self.save(self._model.cfg.save_dir)
            finally:
                self._model.cfg.save_dir = bak
                self._ui_update_done()

    def _on_settings(self):
        cfg_dump_old = self._model.cfg.to_json()

        def _process_cb(apply: bool):
            self._settings_apply = apply
            self._update_from_settings = True
            nonlocal cfg_dump_old
            cfg_dump_new = self._model.cfg.to_json()
            if (cfg_dump_old != cfg_dump_new
                    # not self._model.stage_is_active(self._stage_type_next)
                    or self._model.locs_3d is None):
                cfg_dump_old = cfg_dump_new
                self.stage_start_update()
            else:
                self._show_previews()
                self._settings_apply = False
                self._update_from_settings = False

        self._settings_dlg = FitCfgDialog(self, self._model, process_cb=_process_cb)
        self._settings_dlg.show_modal()
        self._settings_dlg = None
        if (not self._var_preview_occ.get()
                and not self._var_preview_hist.get()
                and not self._var_preview_3d.get()):
            self._settings_apply = False
            self._update_from_settings = False
        self._var_settings.set(0)

    def _on_start(self):
        self._var_start.set(0)

        if self._update_thread is not None:
            if self._update_thread_abort is not None:
                if not self._update_thread_abort.is_set():
                    self._ui_start.configure(state=tk.DISABLED)
                    self._var_summary.set('Aborting...')
                    self._update_thread_abort.set()
        else:
            self._flash_stop()
            self.stage_start_update()

    def _show_previews(self, force_update: bool = False):
        # if not self._model.stage_is_active(self._stage_type_next):
        if self._model.locs_3d is None:
            return

        show = self._model.cfg.show_graphs and self._model.cfg.show_result_graphs

        force_show_occ = show and self._model.cfg.show_all_debug_graphs
        force_update_occ = (force_update
                            and self._model.graph_exists(GraphType.OCCURRENCES))
        if force_show_occ:
            self._var_preview_occ.set(1)
        if force_show_occ or force_update_occ:
            self._on_preview_occ(force_update_occ)

        force_show_hist = show and self._model.cfg.show_all_debug_graphs
        force_update_hist = (force_update
                             and self._model.graph_exists(GraphType.HISTOGRAM))
        if force_show_hist:
            self._var_preview_hist.set(1)
        if force_show_hist or force_update_hist:
            self._on_preview_hist(force_update_hist)

        force_show_3d = show or self._settings_apply
        force_update_3d = (force_update
                           and self._model.graph_exists(GraphType.LOCS_3D))
        if force_show_3d:
            self._var_preview_3d.set(1)
        if force_show_3d or force_update_3d:
            self._on_preview_3d(force_update_3d)

    def _on_preview_occ(self, force_update: bool = False):

        def draw(f: Figure, set_size: bool = True) -> Figure:
            lateral_err = self._model.locs_3d[:, 3]  # Fitting error in X and Y (in microns)
            axial_err = self._model.locs_3d[:, 4]  # Fitting error in Z (in microns)
            photons = self._model.locs_3d[:, 6]  # Number of photons in fit
            return smlfm.graphs.draw_occurrences(
                f,
                lateral_err[self._keep_idx],
                axial_err[self._keep_idx],
                photons[self._keep_idx],
                set_default_size=set_size)

        wnd = self._model.graphs[GraphType.OCCURRENCES]
        if wnd is None:
            fig = draw(Figure())
            wnd = FigureWindow(fig, master=self, title='Occurrences')
            wnd.bind('<Map>', func=lambda _evt: self._var_preview_occ.set(1))
            wnd.bind('<Unmap>', func=lambda _evt: self._var_preview_occ.set(0))
            self._model.graphs[GraphType.OCCURRENCES] = wnd
        else:
            if force_update:
                draw(wnd.figure, set_size=False)
                wnd.refresh()

        if self._var_preview_occ.get():
            wnd.deiconify()
        else:
            wnd.withdraw()

    def _on_preview_hist(self, force_update: bool = False):

        def draw(f: Figure, set_size: bool = True) -> Figure:
            axial_err = self._model.locs_3d[:, 4]  # Fitting error in Z (in microns)
            photons = self._model.locs_3d[:, 6]  # Number of photons in fit
            return smlfm.graphs.draw_histogram(
                f,
                photons[self._keep_idx],
                axial_err[self._keep_idx],
                set_default_size=set_size)

        wnd = self._model.graphs[GraphType.HISTOGRAM]
        if wnd is None:
            fig = draw(Figure())
            wnd = FigureWindow(fig, master=self, title='Histogram')
            wnd.bind('<Map>', func=lambda _evt: self._var_preview_hist.set(1))
            wnd.bind('<Unmap>', func=lambda _evt: self._var_preview_hist.set(0))
            self._model.graphs[GraphType.HISTOGRAM] = wnd
        else:
            if force_update:
                draw(wnd.figure, set_size=False)
                wnd.refresh()

        if self._var_preview_hist.get():
            wnd.deiconify()
        else:
            wnd.withdraw()

    def _on_preview_3d(self, force_update: bool = False):

        def draw(f: Figure, set_size: bool = True) -> Figure:
            xyz = self._model.locs_3d[:, 0:3]  # X, Y, Z
            return smlfm.graphs.draw_3d_locs(
                f,
                xyz[self._keep_idx],
                set_default_size=set_size)

        wnd = self._model.graphs[GraphType.LOCS_3D]
        if wnd is None:
            fig = draw(Figure())
            wnd = FigureWindow(fig, master=self, title='3D')
            wnd.bind('<Map>', func=lambda _evt: self._var_preview_3d.set(1))
            wnd.bind('<Unmap>', func=lambda _evt: self._var_preview_3d.set(0))
            self._model.graphs[GraphType.LOCS_3D] = wnd
        else:
            if force_update:
                draw(wnd.figure, set_size=False)
                wnd.refresh()

        if self._var_preview_3d.get():
            wnd.deiconify()
        else:
            wnd.withdraw()

    # Executed on thread
    def _update_task(self):
        tic = time.time()

        fit_params_all = dataclasses.replace(
            self._model.cfg.fit_params_full,
            frame_min=(self._model.cfg.fit_params_full.frame_min
                       if self._model.cfg.fit_params_full.frame_min > 0
                       else self._model.lfl.min_frame),
            frame_max=(self._model.cfg.fit_params_full.frame_max
                       if self._model.cfg.fit_params_full.frame_max > 0
                       else self._model.lfl.max_frame),
        )

        self._model.invoke_on_gui_thread_async(
            self._var_summary.set,
            f'Fitting frames'
            f' {fit_params_all.frame_min}-{fit_params_all.frame_max}...')

        self._model.locs_3d, _ = smlfm.Fitting.light_field_fit(
            self._model.lfl.corrected_locs_2d,
            self._model.lfm.rho_scaling, fit_params_all,
            abort_event=self._update_thread_abort,
            progress_func=lambda frame, min_frame, _max_frame:
                self._model.invoke_on_gui_thread_async(
                    self._var_progress.set, frame - min_frame + 1),
            progress_step=100,
            worker_count=self._model.cfg.max_workers)

        if self._update_thread_abort is not None:
            if self._update_thread_abort.is_set():
                self._model.locs_3d = None
                return

        lateral_err = self._model.locs_3d[:, 3]  # Fitting error in X and Y (in microns)
        view_count = self._model.locs_3d[:, 5]  # Number of views used to fit the localisation
        max_lateral_err = self._model.cfg.show_max_lateral_err
        min_view_count = self._model.cfg.show_min_view_count
        self._keep_idx = np.logical_and(
            (lateral_err < max_lateral_err) if max_lateral_err is not None else True,
            (view_count > min_view_count) if min_view_count is not None else True)

        if self._model.cfg.log_timing:
            print(f'Complete fitting took {1e3 * (time.time() - tic):.3f} ms')

        if self._update_thread_abort is not None:
            if self._update_thread_abort.is_set():
                self._model.locs_3d = None
                self._keep_idx = None
                return

        if self._model.cfg.save_dir is not None:
            timestamp_str = self._model.save_timestamp.strftime('%Y%m%d-%H%M%S')
            sub_dir = Path(f'3D Fitting - {timestamp_str}'
                           f' - {self._model.cfg.csv_file.name}')
            self._saved_path = self.save(sub_dir)

    # Executed on both, main thread and update thread
    def save(self, folder: Path) -> Path:
        results = smlfm.OutputFiles(self._model.cfg, folder)

        results.mkdir()
        results.save_config()
        results.save_csv(self._model.locs_3d)
        results.save_visp(self._model.locs_3d)
        results.save_figures()

        return results.folder



import os
import tkinter as tk
from idlelib.tooltip import Hovertip
from pathlib import Path
from tkinter import filedialog, ttk
from typing import Callable, Optional

from .app_model import AppModel
from .cfg_dialog import CfgDialog


class ConfigCfgDialog(CfgDialog):

    def __init__(self, parent, model: AppModel, title: Optional[str] = None,
                 process_cb: Optional[Callable[[bool], None]] = None):
        if title is None:
            title = 'Common Settings'

        self._ui_fiji_dir: Optional[ttk.Entry] = None
        fiji_dir_name = '' if model.cfg.fiji_dir is None else str(model.cfg.fiji_dir)
        self._var_fiji_dir = tk.StringVar(value=fiji_dir_name)
        self._ui_fiji_dir_tip: Optional[Hovertip] = None

        self._ui_jvm_opts: Optional[ttk.Entry] = None
        self._var_jvm_opts = tk.StringVar(value=model.cfg.fiji_jvm_opts)

        self._ui_show_graphs: Optional[ttk.Checkbutton] = None
        self._var_show_graphs = tk.BooleanVar(value=model.cfg.show_graphs)

        self._ui_show_result_graphs: Optional[ttk.Checkbutton] = None
        self._var_show_result_graphs = tk.BooleanVar(
            value=model.cfg.show_result_graphs)

        self._ui_show_mla_alignment_graph: Optional[ttk.Checkbutton] = None
        self._var_show_mla_alignment_graph = tk.BooleanVar(
            value=model.cfg.show_mla_alignment_graph)

        self._ui_show_all_debug_graphs: Optional[ttk.Checkbutton] = None
        self._var_show_all_debug_graphs = tk.BooleanVar(
            value=model.cfg.show_all_debug_graphs)

        self._ui_confirm_mla_alignment: Optional[ttk.Checkbutton] = None
        self._var_confirm_mla_alignment = tk.BooleanVar(
            value=model.cfg.confirm_mla_alignment)

        self._ui_log_timing: Optional[ttk.Checkbutton] = None
        self._var_log_timing = tk.BooleanVar(value=model.cfg.log_timing)

        self._ui_max_workers: Optional[ttk.Entry] = None
        self._var_max_workers = tk.StringVar(value=str(model.cfg.max_workers))
        self._ui_max_workers_tip: Optional[Hovertip] = None

        super().__init__(parent, model, title, process_cb=process_cb)

    # noinspection PyProtectedMember
    # pylint: disable=protected-access
    def body(self, master) -> tk.BaseWidget:
        ui_tab = ttk.Frame(master)
        row = 0

        ui_fiji_dir_lbl = ttk.Label(ui_tab, text='Fiji folder:', anchor=tk.E)
        Hovertip(ui_fiji_dir_lbl, text=self.model.cfg._fiji_dir_doc)
        self._ui_fiji_dir = ttk.Entry(ui_tab, textvariable=self._var_fiji_dir)
        self._ui_fiji_dir_tip = Hovertip(
            self._ui_fiji_dir, text=self.model.cfg._fiji_dir_doc)
        ui_fiji_dir_btn = ttk.Button(
            ui_tab, text='...', width=3,
            command=self._on_get_fiji_dir)
        self.ui_widgets.append(self._ui_fiji_dir)
        self.ui_widgets.append(ui_fiji_dir_btn)
        #
        ui_fiji_dir_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_fiji_dir.grid(column=1, row=row, sticky=tk.EW)
        ui_fiji_dir_btn.grid(column=2, row=row, sticky=tk.W)
        row += 1

        ui_jvm_opts_lbl = ttk.Label(ui_tab, text='JVM options:', anchor=tk.E)
        Hovertip(ui_jvm_opts_lbl, text=self.model.cfg._fiji_jvm_opts_doc)
        self._ui_jvm_opts = ttk.Entry(ui_tab, textvariable=self._var_jvm_opts)
        Hovertip(self._ui_jvm_opts, text=self.model.cfg._fiji_jvm_opts_doc)
        self.ui_widgets.append(self._ui_jvm_opts)
        #
        ui_jvm_opts_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_jvm_opts.grid(column=1, row=row, sticky=tk.EW)
        row += 1

        self._ui_show_graphs = ttk.Checkbutton(
            ui_tab, onvalue=True, offvalue=False,
            text='Show graphs',
            variable=self._var_show_graphs)
        Hovertip(self._ui_show_graphs, text=self.model.cfg._show_graphs_doc)
        self.ui_widgets.append(self._ui_show_graphs)
        #
        self._ui_show_graphs.grid(column=0, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        self._ui_show_result_graphs = ttk.Checkbutton(
            ui_tab, onvalue=True, offvalue=False,
            text='Show result graphs',
            variable=self._var_show_result_graphs)
        Hovertip(self._ui_show_result_graphs,
                 text=self.model.cfg._show_result_graphs_doc)
        self.ui_widgets.append(self._ui_show_result_graphs)
        #
        self._ui_show_result_graphs.grid(
            column=0, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        self._ui_show_mla_alignment_graph = ttk.Checkbutton(
            ui_tab, onvalue=True, offvalue=False,
            text='Show MLA alignment',
            variable=self._var_show_mla_alignment_graph)
        Hovertip(self._ui_show_mla_alignment_graph,
                 text=self.model.cfg._show_mla_alignment_graph_doc)
        self.ui_widgets.append(self._ui_show_mla_alignment_graph)
        #
        self._ui_show_mla_alignment_graph.grid(
            column=0, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        self._ui_show_all_debug_graphs = ttk.Checkbutton(
            ui_tab, onvalue=True, offvalue=False,
            text='Show all intermediate graphs',
            variable=self._var_show_all_debug_graphs)
        Hovertip(self._ui_show_all_debug_graphs,
                 text=self.model.cfg._show_all_debug_graphs_doc)
        self.ui_widgets.append(self._ui_show_all_debug_graphs)
        #
        self._ui_show_all_debug_graphs.grid(
            column=0, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        self._ui_confirm_mla_alignment = ttk.Checkbutton(
            ui_tab, onvalue=True, offvalue=False,
            text='Confirm MLA alignment',
            variable=self._var_confirm_mla_alignment)
        Hovertip(self._ui_confirm_mla_alignment,
                 text=self.model.cfg._confirm_mla_alignment_doc)
        self.ui_widgets.append(self._ui_confirm_mla_alignment)
        #
        self._ui_confirm_mla_alignment.grid(
            column=0, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        self._ui_log_timing = ttk.Checkbutton(
            ui_tab, onvalue=True, offvalue=False,
            text='Log timing to console',
            variable=self._var_log_timing)
        Hovertip(self._ui_log_timing, text=self.model.cfg._log_timing_doc)
        self.ui_widgets.append(self._ui_log_timing)
        #
        self._ui_log_timing.grid(column=0, row=row, sticky=tk.EW, columnspan=2)
        row += 1

        max_cpu = os.cpu_count()
        limits = ('\n\nCPU count not detected.' if max_cpu is None else
                  f'\n\nDetected CPU count: {max_cpu}')
        #
        ui_max_workers_lbl = ttk.Label(ui_tab, text='Max. workers:', anchor=tk.E)
        Hovertip(ui_max_workers_lbl, text=self.model.cfg._max_workers_doc + limits)
        self._ui_max_workers = ttk.Spinbox(
            ui_tab, from_=0, to=os.cpu_count(), increment=1, format='%.0f',
            textvariable=self._var_max_workers)
        self._ui_max_workers_tip = Hovertip(
            self._ui_max_workers, text=self.model.cfg._max_workers_doc)
        self.ui_widgets.append(self._ui_max_workers)
        #
        ui_max_workers_lbl.grid(column=0, row=row, sticky=tk.EW)
        self._ui_max_workers.grid(column=1, row=row, sticky=tk.EW)
        row += 1

        # Final padding and layout
        ui_tab.columnconfigure(index=1, weight=1)
        self.model.grid_slaves_configure(ui_tab, padx=1, pady=1)

        ui_tab.pack(anchor=tk.NW, fill=tk.X, expand=True, padx=5, pady=5)

        return self._ui_fiji_dir  # Control that gets initial focus

    def _on_get_fiji_dir(self) -> None:
        dir_name = filedialog.askdirectory(
            parent=self,
            title='Select Fiji.app folder...',
            initialdir=self._var_fiji_dir.get())
        if dir_name:
            self._var_fiji_dir.set(dir_name)

    def validate(self) -> bool:
        fiji_dir = Path(self._var_fiji_dir.get())
        if not fiji_dir.exists() or not fiji_dir.is_dir():
            self.initial_focus = self._ui_fiji_dir
            self._ui_fiji_dir_tip.showtip()
            return False
        if not self.is_int(self._var_max_workers.get()):
            self.initial_focus = self._ui_max_workers
            self._ui_max_workers_tip.showtip()
            return False

        return super().validate()

    def process(self) -> None:
        self.model.cfg.fiji_dir = Path(self._var_fiji_dir.get())
        self.model.cfg.fiji_jvm_opts = self._var_jvm_opts.get()
        self.model.cfg.show_graphs = self._var_show_graphs.get()
        self.model.cfg.show_result_graphs = self._var_show_result_graphs.get()
        self.model.cfg.show_mla_alignment_graph = self._var_show_mla_alignment_graph.get()
        self.model.cfg.show_all_debug_graphs = self._var_show_all_debug_graphs.get()
        self.model.cfg.confirm_mla_alignment = self._var_confirm_mla_alignment.get()
        self.model.cfg.log_timing = self._var_log_timing.get()
        self.model.cfg.max_workers = self.clamp(
            int(self._var_max_workers.get()), 0, os.cpu_count())

        super().process()



from typing import Optional

import numpy.typing as npt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ..graphs import add_watermark


def draw_locs_normalised(fig: Figure,
                         xy: npt.NDArray[float],
                         uv: npt.NDArray[float],
                         lens_idx: npt.NDArray[float],
                         mla_centre: Optional[npt.NDArray[float]] = None,
                         set_default_size: bool = True
                         ) -> Figure:
    fig.clear(True)
    if set_default_size:
        fig.set_size_inches(5, 5)
    fig.set_layout_engine('tight')

    ax: Axes = fig.add_subplot()
    ax.set_xticks([])
    ax.set_yticks([])

    ax.scatter(xy[:, 0], xy[:, 1],
               s=1, c=lens_idx, marker=',', lw=0)

    ax.scatter(uv[:, 0], uv[:, 1],
               s=3, c='tomato')

    if mla_centre is not None:
        ax.scatter(mla_centre[0], mla_centre[1],
                   s=20, c='red', marker='x')

    ax.set_aspect('equal')
    add_watermark(fig)
    return fig



import numpy.typing as npt
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ..graphs import add_watermark


def draw_locs_csv(fig: Figure,
                  locs: npt.NDArray[float],
                  set_default_size: bool = True
                  ) -> Figure:
    fig.clear(True)
    if set_default_size:
        fig.set_size_inches(5, 5)
    fig.set_layout_engine('tight')

    ax: Axes = fig.add_subplot()
    ax.scatter(locs[:, 0], locs[:, 1],
               s=1, c='green', marker=',', lw=0)

    ax.set_aspect('equal')
    add_watermark(fig)
    return fig



from pathlib import Path

import numpy as np
import numpy.typing as npt

from .config import Config


class OutputFiles:

    CONFIG_FILE_NAME: str = 'config.json'
    CSV_FILE_NAME: str = 'locs3D.csv'
    # noinspection SpellCheckingInspection
    VISP_FILE_NAME: str = 'ViSP.3d'
    FIGURES_FILE_NAME: str = 'figures.pyw'

    def __init__(self, cfg: Config, folder: Path):
        self.cfg = cfg
        if folder.is_absolute():
            self.folder = folder
        else:
            self.folder = cfg.save_dir / folder

    def mkdir(self, mode=0o777):
        # Like POSIX 'mkdir -p'
        self.folder.mkdir(mode=mode, parents=True, exist_ok=True)

    def save_config(self):
        cfg_dump = self.cfg.to_json()
        with open(self.folder / self.CONFIG_FILE_NAME, 'wt', encoding='utf-8') as file:
            file.write(cfg_dump)

    def save_csv(self, locs_3d: npt.NDArray[float]):
        np.savetxt(self.folder / self.CSV_FILE_NAME, locs_3d,
                   fmt='%.17s,%.17s,%.17s,%.17s,%.17s,%d,%.17s,%d')

    # noinspection SpellCheckingInspection
    def save_visp(self, locs_3d: npt.NDArray[float]):
        xyz = locs_3d[:, 0:3]
        photons = locs_3d[:, 6]
        frames = locs_3d[:, 7]
        visp = np.column_stack([xyz * 1000, photons, frames])
        np.savetxt(self.folder / self.VISP_FILE_NAME, visp,
                   fmt='%.8s\t%.8s\t%.8s\t%.8s\t%d')

    def save_figures(self):
        # Generate .pyw file that shows all figures (relies on the package)
        # noinspection SpellCheckingInspection
        output_str = (
            f'#!/usr/bin/env python3\n'
            f'# # # GENERATED FILE # # #\n'
            f'from pathlib import Path\n'
            f'from matplotlib import pyplot as plt\n'
            f'import smlfm\n'
            f'if __name__ == "__main__":\n'
            f'    locs_3d = Path(__file__).parent.absolute() / "{self.CSV_FILE_NAME}"\n'
            f'    max_lat_err = {self.cfg.show_max_lateral_err}\n'
            f'    min_views = {self.cfg.show_min_view_count}\n'
            f'    fig1, fig2, fig3 = smlfm.graphs.reconstruct_results(\n'
            f'        plt.figure(), plt.figure(), plt.figure(),\n'
            f'        locs_3d, max_lat_err, min_views)\n'
            f'    fig1.canvas.manager.set_window_title("Occurrences")\n'
            f'    fig2.canvas.manager.set_window_title("Histogram")\n'
            f'    fig3.canvas.manager.set_window_title("3D")\n'
            f'    plt.show()\n'
        )
        with open(self.folder / self.FIGURES_FILE_NAME, 'wt', encoding='utf-8') as file:
            file.write(output_str)



from enum import Enum, unique

import numpy as np
import numpy.typing as npt


class MicroLensArray:
    """A class describing micro-lens array.

    Attributes:
        lattice_type (LatticeType): A lens arrangement.
        focal_length (float): A focal length of micro-lenses (in mm).
        lens_pitch (float):
            A pitch along X direction, parallel with optical table (in microns).
        optic_size (float):
            A physical size of the array, assumes a square optic (in microns).
        centre (npt.NDArray[float]):
            A spatial X,Y position in terms of camera space
            (in lattice spacing units).

        lens_centres (npt.NDArray[float]):
            An array of X,Y lens centres (in lattice spacing units).
    """

    @unique
    class LatticeType(Enum):
        HEXAGONAL = 1
        SQUARE = 2

    def __init__(self,
                 lattice_type: LatticeType,
                 focal_length: float,
                 lens_pitch: float,
                 optic_size: float,
                 centre: npt.NDArray[float]):
        """Constructs the micro-lens array object.

        Args:
            lattice_type (LatticeType): Lens arrangement.
            focal_length (float): A focal length of micro-lenses (in mm).
            lens_pitch (float):
                A pitch along X direction, parallel with optical table
                (in microns).
            optic_size (float):
                A physical size of the array, assumes it is a square optic
                (in microns).
            centre (npt.NDArray[float]):
                A spatial X, Y position in terms of camera space
                (in lattice spacing units).
        """

        self.lattice_type = lattice_type
        self.focal_length = focal_length
        self.lens_pitch = lens_pitch
        self.optic_size = optic_size
        self.centre = centre

        self.lens_centres = self._generate_lattice()

    def _generate_lattice(self) -> npt.NDArray[float]:
        """Generate the X,Y lattice coordinates based on given pattern."""

        if self.lattice_type == MicroLensArray.LatticeType.SQUARE:
            # Generate centers of square lenses, pretending pitch is 1.
            # Generate more than needed, because gets cropped when rotated.

            width = self.optic_size / self.lens_pitch
            marks = np.arange(-np.floor(width / 2), np.ceil(width / 2) + 1)
            x, y = np.meshgrid(marks, marks)

        elif self.lattice_type == MicroLensArray.LatticeType.HEXAGONAL:
            # Generate centers of hexagonal lenses, pretending pitch is 1.
            # Generate more than needed, because gets cropped when rotated.

            width = self.optic_size / self.lens_pitch
            marks = np.arange(-np.floor(width / 2), np.ceil(width / 2) + 1)
            xx, yy = np.meshgrid(marks, marks)

            length = np.max(xx.shape)
            dx = np.tile([0.5, 0], [length, np.ceil(length / 2).astype(int)])
            dx = dx[0:length, 0:length]  # TODO: Is this necessary?
            x = yy * np.sqrt(3) / 2
            y = xx + dx.T + 0.5

        else:
            raise ValueError(f'Unsupported lattice type with value {self.lattice_type}')

        return np.column_stack((x.flatten('F'), y.flatten('F')))

    def offset_lattice(self,
                       dxy: npt.NDArray[float]
                       ) -> None:
        """Shift the coordinates of the micro-lenses centers.

        Args:
            dxy (npt.NDArray[float]): A shift in X and Y directions
                (in lattice spacing units).
        """

        self.lens_centres[:] += dxy

        # TODO: Should the MLA centre be moved as well?
        #       If yes, it negates the move effect...
        # self.centre += dxy

    def rotate_lattice(self,
                       theta: float
                       ) -> None:
        """Rotate the coordinates of the micro-lenses centers around the centre.

        Args:
            theta (float): An angle to rotate (in radians).
        """

        # Shift centre to origin
        x = self.lens_centres[:, 0] - self.centre[0]
        y = self.lens_centres[:, 1] - self.centre[1]
        # Rotate around origin
        self.lens_centres[:, 0] = (x * np.cos(theta) - y * np.sin(theta))
        self.lens_centres[:, 1] = (x * np.sin(theta) + y * np.cos(theta))
        # Shift origin back to centre
        self.lens_centres[:, 0] += self.centre[0]
        self.lens_centres[:, 1] += self.centre[1]



from typing import Optional

import numpy.typing as npt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.patches import Circle

from ..graphs import add_watermark


def draw_mla(fig: Figure,
             lens_centres: npt.NDArray[float],
             mla_centre: Optional[npt.NDArray[float]] = None,
             bfp_radius: Optional[float] = None,
             set_default_size: bool = True
             ) -> Figure:
    fig.clear(True)
    if set_default_size:
        fig.set_size_inches(5, 5)
    fig.set_layout_engine('tight')

    ax: Axes = fig.add_subplot()

    if bfp_radius is not None and mla_centre is not None:
        bfp = Circle(xy=(float(mla_centre[0]), float(mla_centre[1])),
                     radius=bfp_radius, fill=False,
                     edgecolor='purple', alpha=0.25)
        ax.add_patch(bfp)

    ax.scatter(lens_centres[:, 0], lens_centres[:, 1],
               s=3, c='tomato', marker='o')

    if mla_centre is not None:
        ax.scatter(mla_centre[0], mla_centre[1],
                   s=20, c='red', marker='x')

    for i in range(lens_centres.shape[0]):
        ax.annotate(str(i),
                    xy=(lens_centres[i, 0], lens_centres[i, 1]),
                    xycoords='data',
                    xytext=(1.5, 1.0),
                    alpha=0.5,
                    textcoords='offset points',
                    size=10)

    ax.set_aspect('equal')
    add_watermark(fig)
    return fig
